import base64
import json
import os
import subprocess
import tempfile
import time
import traceback
import click_spinner
import click
from typing import List, Optional, Dict

import yaml

import typer

from ..cli.utils import (
    log_title,
    fetch_runner_logs,
    exec_in_robusta_runner,
    namespace_to_kubectl,
    PLAYBOOKS_DIR, get_package_name,
)

PLAYBOOKS_MOUNT_LOCATION = "/etc/robusta/playbooks/storage"

NAMESPACE_EXPLANATION = (
    "Installation namespace. If none use the namespace currently active with kubectl."
)
CONFIG_SECRET_NAME = "robusta-playbooks-config-secret"

app = typer.Typer()


def get_runner_pod(namespace: str) -> Optional[str]:
    pods = subprocess.check_output(
        f'kubectl get pods {namespace_to_kubectl(namespace)} --selector="robustaComponent=runner"',
        shell=True,
    )
    text = pods.decode("utf-8")
    if not text:
        return None

    lines = text.split("\n")
    if len(lines) < 2:
        return None
    return lines[1].split(" ")[0]


def __validate_playbooks_dir(playbooks_dir: str) -> bool:
    files = os.listdir(playbooks_dir)
    if "pyproject.toml" not in files:
        typer.secho("Illegal playbooks package. missing pyproject.toml", fg="red")
        return False

    package_name = get_package_name(os.path.join(playbooks_dir))
    if not package_name:
        typer.secho("Illegal pyproject.toml. `name` attribute not defined in pyproject.toml", fg="red")
        return False

    if package_name not in files:
        typer.secho(f"Playbooks directory missing or does not match playbooks package name {package_name}", fg="red")
        return False

    return True


@app.command()
def push(
    playbooks_directory: str = typer.Argument(
        ...,
        help="Local playbooks code directory",
    ),
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):
    """Load custom playbooks code"""
    log_title("Uploading playbooks code...")
    with fetch_runner_logs(namespace):
        runner_pod = get_runner_pod(namespace)
        if not runner_pod:
            log_title(f"Runner pod not found in the {namespace} namespace. If robusta is installed in a different namespace, use the --namespace flag.", color="red")
            return

        subprocess.check_call(
            f"kubectl exec -it {namespace_to_kubectl(namespace)} {runner_pod} -c runner "
            f"-- bash -c 'mkdir -p {PLAYBOOKS_MOUNT_LOCATION}'",
            shell=True,
        )
        abs_path = os.path.abspath(playbooks_directory)
        dir_name = os.path.basename(os.path.normpath(abs_path))
        if not __validate_playbooks_dir(abs_path):
            return

        subprocess.check_call(
            f"kubectl cp {namespace_to_kubectl(namespace)} {abs_path} "
            f"{runner_pod}:{PLAYBOOKS_MOUNT_LOCATION}/{dir_name} -c runner",
            shell=True,
        )
        time.sleep(
            5
        )  # wait five seconds for the runner to actually reload the playbooks
    log_title("Loaded custom playbooks code!")


@app.command()
def configure(
    config_file: str = typer.Argument(
        ...,
        help="Playbooks configuration file",
    ),
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):
    """Deploy playbooks configuration"""
    log_title("Configuring playbooks...")
    with fetch_runner_logs(namespace):
        subprocess.check_call(
            f"kubectl create secret generic {namespace_to_kubectl(namespace)} {CONFIG_SECRET_NAME} "
            f"--from-file active_playbooks.yaml={config_file} --type=Opaque -o yaml --dry-run | kubectl apply -f -",
            shell=True,
        )
        subprocess.check_call(
            f'kubectl annotate pods {namespace_to_kubectl(namespace)} -l robustaComponent=runner '
            f'--overwrite "playbooks-last-modified={time.time()}"',
            shell=True,
        )
        time.sleep(
            5
        )  # wait five seconds for the runner to actually reload the playbooks
    log_title("Deployed playbooks!")


def get_playbooks_config(namespace: str):
    configmap_content = subprocess.check_output(
        f"kubectl get secret {namespace_to_kubectl(namespace)} {CONFIG_SECRET_NAME} -o yaml",
        shell=True,
    )
    playbooks_secret = yaml.safe_load(configmap_content)
    playbooks_secret["data"]["active_playbooks.yaml"] = base64.b64decode(
        playbooks_secret["data"]["active_playbooks.yaml"]
    ).decode()
    return playbooks_secret


@app.command()
def pull(
    playbooks_directory: str = typer.Argument(
        ...,
        help="Local target directory",
    ),
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):
    """pull cluster deployed playbooks"""
    if not playbooks_directory:
        playbooks_directory = os.path.join(os.getcwd(), PLAYBOOKS_DIR)

    log_title(f"Pulling playbooks into {playbooks_directory} ")

    try:
        runner_pod = get_runner_pod(namespace)
        if not runner_pod:
            log_title("Runner pod not found.", color="red")
            return

        subprocess.check_call(
            f"kubectl cp {namespace_to_kubectl(namespace)} "
            f"{runner_pod}:{PLAYBOOKS_MOUNT_LOCATION}/ -c runner {playbooks_directory}",
            shell=True,
        )
    except Exception as e:
        typer.echo(f"Failed to pull deployed playbooks {traceback.format_exc()}")


@app.command("list-dirs")
def list_dirs(
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):
    """List stored playbooks directories"""
    log_title(f"Listing playbooks directories ")

    try:
        runner_pod = get_runner_pod(namespace)
        if not runner_pod:
            log_title("Runner pod not found.", color="red")
            return

        ls_res = subprocess.check_output(
            f"kubectl exec -it {namespace_to_kubectl(namespace)} {runner_pod} -c runner "
            f"-- bash -c 'ls {PLAYBOOKS_MOUNT_LOCATION}'",
            shell=True,
        )

        log_title(f"Stored playbooks directories: \n { ls_res.decode('utf-8')}")

    except Exception as e:
        typer.echo(f"Failed to list deployed playbooks {traceback.format_exc()}")


@app.command()
def delete(
    playbooks_directory: str = typer.Argument(
        ...,
        help="Playbooks directory that should be deleted",
    ),
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):
    """delete playbooks directory from storage"""
    if not playbooks_directory:
        log_title("Playbooks directory not specified", "red")
        return

    log_title(f"Deleting playbooks directory {playbooks_directory} ")

    try:
        runner_pod = get_runner_pod(namespace)
        if not runner_pod:
            log_title("Runner pod not found.", color="red")
            return

        path_to_delete = os.path.join(PLAYBOOKS_MOUNT_LOCATION, playbooks_directory)
        subprocess.check_call(
            f"kubectl exec -it {namespace_to_kubectl(namespace)} {runner_pod} -c runner "
            f"-- bash -c 'rm -rf {path_to_delete}'",
            shell=True,
        )

    except Exception as e:
        typer.echo(f"Failed to delete deployed playbooks {traceback.format_exc()}")


def print_yaml_if_not_none(key: str, json_dict: dict):
    if json_dict.get(key):
        json = {}
        json[key] = json_dict.get(key)
        typer.echo(f"{yaml.dump(json)}")


@app.command("list")
def list_(
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):  # not named list as that would shadow the builtin list function
    """list current active playbooks"""
    typer.echo(f"Getting deployed playbooks list...")
    with click_spinner.spinner():
        playbooks_config = get_playbooks_config(namespace)

    active_playbooks_file = playbooks_config["data"]["active_playbooks.yaml"]
    active_playbooks_yaml = yaml.safe_load(active_playbooks_file)
    for playbook in active_playbooks_yaml["active_playbooks"]:
        typer.secho(f"--------------------------------------", fg="blue")
        print_yaml_if_not_none("sinks", playbook)
        print_yaml_if_not_none("triggers", playbook)
        print_yaml_if_not_none("actions", playbook)


@app.command()
def edit_config(
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
):
    """show and edit active_playbooks.yaml from cluster"""
    typer.echo("connecting to cluster...")
    with click_spinner.spinner():
        playbooks_config = get_playbooks_config(namespace)
    active_playbooks_file = playbooks_config["data"]["active_playbooks.yaml"]
    edited_result = click.edit(active_playbooks_file)
    if edited_result is None:
        typer.echo("file not saved in editor. nothing to update")
    elif edited_result.strip() == active_playbooks_file.strip():
        typer.echo("saved file is the same. nothing to update")
    else:
        typer.echo("file modified; updating server")
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(edited_result.encode())
            f.flush()
            fname = f.name
        configure(config_file=fname, namespace=namespace)


def _post_in_runner_pod(namespace: str, api_path: str, req_body: Dict, req_name: str):
    with fetch_runner_logs(namespace=namespace):
        cmd = (
            f"curl -X POST http://localhost:5000/api/{api_path} "
            f"-H 'Content-Type: application/json' "
            f"-d '{json.dumps(req_body)}'"
        )
        exec_in_robusta_runner(
            cmd,
            namespace=namespace,
            tries=3,
            error_msg=f"Cannot {req_name} - usually this means Robusta just started. Will try again",
        )
        typer.echo("\n")


@app.command()
def trigger(
    action_name: str,
    param: Optional[List[str]] = typer.Argument(
        None,
        help="data to send to action (can be used multiple times)",
        metavar="key=value",
    ),
    namespace: str = typer.Option(
        None,
        help="Robusta namespace",
    ),
):
    """trigger a manually run playbook"""
    log_title("Triggering action...")
    action_params = {}
    for p in param:
        (key, val) = p.split("=")
        action_params[key] = val
    # action_params = " ".join([f"-F '{p}'" for p in param])
    req_body = {"action_name": action_name, "action_params": action_params}

    _post_in_runner_pod(namespace=namespace, api_path="trigger", req_body=req_body, req_name="trigger action")
    log_title("Done!")


@app.command()
def reload(
        namespace: str = typer.Option(
            None,
            help="Robusta namespace",
        ),
):
    """reload playbooks configuration"""
    log_title("Reloading playbooks...")
    _post_in_runner_pod(namespace=namespace, api_path="playbooks/reload", req_body={}, req_name="reload playbooks")
    log_title("Done!")


if __name__ == "__main__":
    app()



import logging
import time
from threading import Thread, Lock
from queue import Queue, Full
import prometheus_client

from robusta.core.model.env_vars import INCOMING_EVENTS_QUEUE_MAX_SIZE


class QueueMetrics:
    def __init__(self):
        self.queued = prometheus_client.Counter(
            "queued", "Number of queued events", labelnames=("queue_name",)
        )
        self.processed = prometheus_client.Counter(
            "processed", "Number of processed events", labelnames=("queue_name",)
        )
        self.rejected = prometheus_client.Counter(
            "rejected", "Number of rejected events", labelnames=("queue_name",)
        )
        self.total_process_time = prometheus_client.Summary(
            "total_process_time",
            "Total process time (seconds)",
            labelnames=("queue_name",),
        )

    def on_rejected(self, queue_name):
        self.rejected.labels([queue_name]).inc()

    def on_queued(self, queue_name):
        self.queued.labels([queue_name]).inc()

    def on_processed(self, queue_name, processing_time: float):
        self.processed.labels([queue_name]).inc()
        self.total_process_time.labels([queue_name]).observe(processing_time)


class TaskQueue(Queue):
    def __init__(self, name: str, num_workers, metrics: QueueMetrics):
        Queue.__init__(self, maxsize=INCOMING_EVENTS_QUEUE_MAX_SIZE)
        logging.info(
            f"Initialized task queue: {num_workers} workers. Max size {INCOMING_EVENTS_QUEUE_MAX_SIZE}"
        )
        self.name = name
        self.num_workers = num_workers
        self.metrics = metrics
        self.__start_workers()

    def add_task(self, task, *args, **kwargs):
        args = args or ()
        kwargs = kwargs or {}
        try:
            self.put((task, args, kwargs), block=False)
            self.metrics.on_queued(self.name)
        except Full:
            self.metrics.on_rejected(self.name)

    def __start_workers(self):
        for i in range(self.num_workers):
            t = Thread(target=self.worker)
            t.daemon = True
            t.start()

    def worker(self):
        while True:
            item, args, kwargs = self.get()
            start_time = time.time()
            item(*args, **kwargs)
            self.metrics.on_processed(self.name, time.time() - start_time)
            self.task_done()



from pydantic import PrivateAttr

from .base_trigger import BaseTrigger
from ..triggers.custom_triggers import CustomTriggers
from ...integrations.prometheus.trigger import PrometheusAlertTriggers
from ...integrations.kubernetes.autogenerated.triggers import K8sTriggers
from ...integrations.scheduled.trigger import ScheduledTriggers


class Trigger(K8sTriggers, PrometheusAlertTriggers, ScheduledTriggers, CustomTriggers):
    _trigger: BaseTrigger = PrivateAttr()

    def __init__(self, *args, **data):
        super().__init__(*args, **data)
        trigger_keys = [trigger for trigger in dir(self) if not trigger.startswith("_")]
        triggers = []
        for key in trigger_keys:
            trigger_def = getattr(self, key)
            if trigger_def is not None and isinstance(trigger_def, BaseTrigger):
                triggers.append(trigger_def)
        if len(triggers) != 1:
            raise Exception("Exactly one trigger type must be defined")
        self._trigger = triggers[0]

    def get(self) -> BaseTrigger:
        return self._trigger



from robusta.api import *
from robusta.core.model.services import ServiceInfo
from robusta.core.discovery.top_service_resolver import TopServiceResolver

from robusta.core.playbooks.common import get_events_list, get_event_timestamp

@action
def cluster_discovery_updates(event: KubernetesAnyChangeEvent):
    if (
        event.operation in [K8sOperationType.CREATE, K8sOperationType.UPDATE]
        and event.obj.kind in ["Deployment", "ReplicaSet", "DaemonSet", "StatefulSet"]
        and not event.obj.metadata.ownerReferences
    ):
        TopServiceResolver.add_cached_service(
            ServiceInfo(
                name=event.obj.metadata.name,
                service_type=event.obj.kind,
                namespace=event.obj.metadata.namespace,
            )
        )


@action
def event_history(event: ExecutionBaseEvent):
    """
    Creates findings for the all the past events for any object that has had at one point a warning event
    """
    reported_obj_history_list = []
    warning_events = get_events_list(event_type="Warning")
    for warning_event in warning_events.items:
        warning_event_key = gen_object_key(warning_event.involvedObject)
        if warning_event_key in reported_obj_history_list:
            # if there were multiple warnings on the same object we dont want the history pulled multiple times
            continue
        finding = create_debug_event_finding(warning_event)
        events_table = get_resource_events_table(
            "Resource events",
            warning_event.involvedObject.kind,
            warning_event.involvedObject.name,
            warning_event.involvedObject.namespace
        )
        if events_table:
            finding.add_enrichment([events_table])
            event.add_finding(finding)
        reported_obj_history_list.append(warning_event_key)


def create_debug_event_finding(event: Event):
    """
    Create finding based on the kubernetes event
    """
    k8s_obj = event.involvedObject

    finding = Finding(
        title=f"{event.reason} {event.type} for {k8s_obj.kind} {k8s_obj.namespace}/{k8s_obj.name}",
        description=event.message,
        source=FindingSource.KUBERNETES_API_SERVER,
        severity=FindingSeverity.DEBUG,
        finding_type=FindingType.ISSUE,
        aggregation_key=f"Kubernetes {event.type} Event",
        subject=FindingSubject(
            k8s_obj.name,
            FindingSubjectType.from_kind(k8s_obj.kind.lower()),
            k8s_obj.namespace,
        ),
        creation_date= get_event_timestamp(event)
    )
    finding.service_key = TopServiceResolver.guess_service_key(name=k8s_obj.name, namespace=k8s_obj.namespace)
    return finding


def gen_object_key(object):
    return f"{object.namespace}/{object.kind}/{object.name}"



from .sink_base import SinkBase
from .sink_config import SinkConfigBase
from .slack.slack_sink import SlackSink
from .datadog.datadog_sink import DataDogSink
from .kafka.kafka_sink import KafkaSink
from .msteams.msteams_sink import MsTeamsSink
from .robusta.robusta_sink import RobustaSink
from .slack.slack_sink_params import SlackSinkConfigWrapper
from .datadog.datadog_sink_params import DataDogSinkConfigWrapper
from .kafka.kafka_sink_params import KafkaSinkConfigWrapper
from .msteams.msteams_sink_params import MsTeamsSinkConfigWrapper
from .robusta.robusta_sink_params import RobustaSinkConfigWrapper
from .opsgenie.opsgenie_sink import OpsGenieSink
from .opsgenie.opsgenie_sink_params import OpsGenieSinkConfigWrapper
from .telegram.telegram_sink import TelegramSink
from .telegram.telegram_sink_params import TelegramSinkConfigWrapper
from .webhook.webhook_sink import WebhookSink
from .webhook.webhook_sink_params import WebhookSinkConfigWrapper


class SinkFactory:
    @classmethod
    def create_sink(cls, sink_config: SinkConfigBase, registry) -> SinkBase:
        if isinstance(sink_config, SlackSinkConfigWrapper):
            return SlackSink(sink_config, registry)
        elif isinstance(sink_config, RobustaSinkConfigWrapper):
            return RobustaSink(sink_config, registry)
        elif isinstance(sink_config, MsTeamsSinkConfigWrapper):
            return MsTeamsSink(sink_config, registry)
        elif isinstance(sink_config, KafkaSinkConfigWrapper):
            return KafkaSink(sink_config, registry)
        elif isinstance(sink_config, DataDogSinkConfigWrapper):
            return DataDogSink(sink_config, registry)
        elif isinstance(sink_config, OpsGenieSinkConfigWrapper):
            return OpsGenieSink(sink_config, registry)
        elif isinstance(sink_config, TelegramSinkConfigWrapper):
            return TelegramSink(sink_config, registry)
        elif isinstance(sink_config, WebhookSinkConfigWrapper):
            return WebhookSink(sink_config, registry)
        else:
            raise Exception(f"Sink not supported {type(sink_config)}")



import datetime
import logging
import os
import re
import time
import traceback
import io
import tempfile
import tarfile
from typing import List, Optional

from kubernetes import config
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream

from hikaru.model import Job

RUNNING_STATE = "Running"
SUCCEEDED_STATE = "Succeeded"

try:
    if os.getenv("KUBERNETES_SERVICE_HOST"):
        config.load_incluster_config()
    else:
        config.load_kube_config()
except config.config_exception.ConfigException as e:
    logging.warning(f"Running without kube-config! e={e}")


default_exec_command = ["/bin/sh", "-c"]


def wait_until(
    read_function, predicate_function, timeout_sec: float, backoff_wait_sec: float
):
    """
    repeatedly calls predicate_function(read_function)) until predicate_function returns True or we timeout
    return the last result of read_function() on success and raises an exception on timeout
    between attempts, we wait backoff_wait_sec seconds
    """
    start_time_sec = time.time()

    while start_time_sec + timeout_sec > time.time():
        try:
            resp = read_function()
            if predicate_function(resp):
                return resp
        except ApiException as e:
            logging.error(f"failed calling read_function {traceback.format_exc()}")

        time.sleep(backoff_wait_sec)

    raise Exception("Failed to reach wait condition")


def wait_until_job_complete(job: Job, timeout):
    """
    wait until a kubernetes Job object either succeeds or fails at least once
    """

    def is_job_complete(j: Job) -> bool:
        return j.status.completionTime is not None or j.status.failed is not None

    return wait_until(
        lambda: Job.readNamespacedJob(job.metadata.name, job.metadata.namespace).obj,
        is_job_complete,
        timeout,
        5,
    )


# TODO: refactor to use wait_until function
def wait_for_pod_status(
    name, namespace, status: str, timeout_sec: float, backoff_wait_sec: float
) -> str:
    pod_details = f"pod status: {name} {namespace} {status} {timeout_sec}"
    logging.debug(f"waiting for {pod_details}")

    start_time_sec = time.time()
    while start_time_sec + timeout_sec > time.time():
        try:
            core_v1 = core_v1_api.CoreV1Api()
            resp = core_v1.read_namespaced_pod_status(name, namespace)

            if resp.status.phase == status:
                logging.debug(f"reached {pod_details}")
                return status

        except ApiException as e:
            logging.error(
                f"failed to get pod status {name} {namespace} {traceback.format_exc()}"
            )

        time.sleep(backoff_wait_sec)

    logging.debug(f"failed to reach {pod_details}")
    return "FAIL"


def exec_shell_command(name, shell_command: str, namespace="default", container=None):
    commands = default_exec_command.copy()
    commands.append(shell_command)
    return exec_commands(name, commands, namespace, container)


def upload_file(
    name: str, destination: str, contents: bytes, namespace="default", container=None
):
    core_v1 = core_v1_api.CoreV1Api()
    resp = stream(
        core_v1.connect_get_namespaced_pod_exec,
        name,
        namespace,
        container=container,
        command=["tar", "xvf", "-", "-C", "/"],
        stderr=True,
        stdin=True,
        stdout=True,
        tty=False,
        _preload_content=False,
    )

    with tempfile.TemporaryFile() as local_tar_file:
        with tarfile.open(fileobj=local_tar_file, mode="w") as tar:
            tarinfo = tarfile.TarInfo(destination)
            tarinfo.size = len(contents)
            tar.addfile(tarinfo, fileobj=io.BytesIO(contents))

        local_tar_file.seek(0)
        resp.write_stdin(local_tar_file.read())

        while resp.is_open():
            resp.update(timeout=1)
            if resp.peek_stdout():
                print("STDOUT: %s" % resp.read_stdout())
            if resp.peek_stderr():
                print("STDERR: %s" % resp.read_stderr())
            else:
                break
        resp.close()


def get_pod_logs(
    name,
    namespace="default",
    container="",
    previous=None,
    tail_lines=None,
    since_seconds=None,
):
    resp = None
    try:
        core_v1 = core_v1_api.CoreV1Api()
        resp = core_v1.read_namespaced_pod_log(
            name,
            namespace,
            container=container,
            previous=previous,
            tail_lines=tail_lines,
            since_seconds=since_seconds,
            _preload_content=False  # If this flag is not used, double quotes in json objects in stdout are converted
            # by the kubernetes client to single quotes, which makes json.loads() fail.
            # We therefore use this flag in order to get the raw bytes and decode the output
        ).data.decode("utf-8")

    except ApiException as e:
        if e.status != 404:
            logging.exception(f"failed to get pod logs {name} {namespace} {container}")
            resp = ""

    logging.debug(f"get logs {resp}")
    return resp


def prepare_pod_command(cmd) -> Optional[List[str]]:
    if type(cmd) == list:
        return cmd
    elif type(cmd) == str:
        return cmd.split(" ")  # cmd need to be a list of strings
    elif cmd is None:
        return None
    else:
        logging.exception(f"cmd {cmd} has unknown type {type(cmd)}")
        return cmd


def exec_commands(name, exec_command, namespace="default", container=None):
    logging.debug(
        f"Executing command name: {name} command: {exec_command} namespace: {namespace} container: {container}"
    )
    response_stdout = None

    # verify pod state before connecting
    pod_status = wait_for_pod_status(
        name, namespace, RUNNING_STATE, 90, 0.2
    )  # TODO config
    if pod_status != RUNNING_STATE:
        msg = (
            f"Not running exec commands. Pod {name} {namespace} is not in running state"
        )
        logging.error(msg)
        return msg

    wsclient = None
    try:
        core_v1 = core_v1_api.CoreV1Api()
        wsclient = stream(
            core_v1.connect_get_namespaced_pod_exec,
            name,
            namespace,
            container=container,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False,  # fix https://github.com/kubernetes-client/python/issues/811
        )
        wsclient.run_forever()
        response = wsclient.read_all()
        logging.debug(f"exec command response {response}")
    finally:
        if wsclient is not None:
            wsclient.close()

    return response


def to_kubernetes_name(name, prefix=""):
    """
    Returns a valid and unique kubernetes name based on prefix and name, replacing characters in name as necessary
    see https://kubernetes.io/docs/concepts/overview/working-with-objects/names/
    """
    unique_id = str(time.time()).replace(".", "-")
    safe_name = re.sub("[^0-9a-zA-Z\\-]+", "-", name)
    return f"{prefix}{safe_name}-{unique_id}"[:63]


def parse_kubernetes_datetime(k8s_datetime: str) -> datetime.datetime:
    """
    Parse a date/time related field in a Kubernetes object and convert it from string to a python datetime object
    """
    return datetime.datetime.strptime(k8s_datetime, "%Y-%m-%dT%H:%M:%S%z")


def parse_kubernetes_datetime_with_ms(k8s_datetime: str) -> datetime.datetime:
    return datetime.datetime.strptime(k8s_datetime, "%Y-%m-%dT%H:%M:%S.%f%z")


def parse_kubernetes_datetime_to_ms(k8s_datetime: str) -> float:
    """
        for timestamps eventTime has milliseconds and firstTimestamp,lastTimestamp,creationTimestamp do not.
        we parse the more commonly used timestamp first
    """
    try:
        return parse_kubernetes_datetime(k8s_datetime).timestamp() * 1000
    except:
        return parse_kubernetes_datetime_with_ms(k8s_datetime).timestamp() * 1000



# This file was autogenerated. Do not edit.

from typing import Optional, Dict
from pydantic import BaseModel
from ..base_triggers import K8sBaseTrigger
from ....core.model.k8s_operation_type import K8sOperationType
from .events import *


# Pod Triggers
class PodAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Pod", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PodChangeEvent


class PodCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Pod", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PodChangeEvent


class PodDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Pod", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PodChangeEvent


class PodUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Pod", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PodChangeEvent


# ReplicaSet Triggers
class ReplicaSetAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ReplicaSet", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ReplicaSetChangeEvent


class ReplicaSetCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ReplicaSet", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ReplicaSetChangeEvent


class ReplicaSetDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ReplicaSet", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ReplicaSetChangeEvent


class ReplicaSetUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ReplicaSet", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ReplicaSetChangeEvent


# DaemonSet Triggers
class DaemonSetAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="DaemonSet", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DaemonSetChangeEvent


class DaemonSetCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="DaemonSet", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DaemonSetChangeEvent


class DaemonSetDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="DaemonSet", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DaemonSetChangeEvent


class DaemonSetUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="DaemonSet", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DaemonSetChangeEvent


# Deployment Triggers
class DeploymentAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Deployment", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DeploymentChangeEvent


class DeploymentCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Deployment", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DeploymentChangeEvent


class DeploymentDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Deployment", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DeploymentChangeEvent


class DeploymentUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Deployment", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return DeploymentChangeEvent


# StatefulSet Triggers
class StatefulSetAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="StatefulSet", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return StatefulSetChangeEvent


class StatefulSetCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="StatefulSet", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return StatefulSetChangeEvent


class StatefulSetDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="StatefulSet", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return StatefulSetChangeEvent


class StatefulSetUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="StatefulSet", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return StatefulSetChangeEvent


# Service Triggers
class ServiceAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Service", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceChangeEvent


class ServiceCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Service", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceChangeEvent


class ServiceDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Service", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceChangeEvent


class ServiceUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Service", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceChangeEvent


# Event Triggers
class EventAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Event", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return EventChangeEvent


class EventCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Event", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return EventChangeEvent


class EventDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Event", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return EventChangeEvent


class EventUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Event", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return EventChangeEvent


# HorizontalPodAutoscaler Triggers
class HorizontalPodAutoscalerAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="HorizontalPodAutoscaler", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return HorizontalPodAutoscalerChangeEvent


class HorizontalPodAutoscalerCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="HorizontalPodAutoscaler", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return HorizontalPodAutoscalerChangeEvent


class HorizontalPodAutoscalerDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="HorizontalPodAutoscaler", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return HorizontalPodAutoscalerChangeEvent


class HorizontalPodAutoscalerUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="HorizontalPodAutoscaler", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return HorizontalPodAutoscalerChangeEvent


# Node Triggers
class NodeAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Node", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NodeChangeEvent


class NodeCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Node", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NodeChangeEvent


class NodeDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Node", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NodeChangeEvent


class NodeUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Node", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NodeChangeEvent


# ClusterRole Triggers
class ClusterRoleAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRole", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleChangeEvent


class ClusterRoleCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRole", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleChangeEvent


class ClusterRoleDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRole", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleChangeEvent


class ClusterRoleUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRole", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleChangeEvent


# ClusterRoleBinding Triggers
class ClusterRoleBindingAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRoleBinding", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleBindingChangeEvent


class ClusterRoleBindingCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRoleBinding", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleBindingChangeEvent


class ClusterRoleBindingDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRoleBinding", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleBindingChangeEvent


class ClusterRoleBindingUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ClusterRoleBinding", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ClusterRoleBindingChangeEvent


# Job Triggers
class JobAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Job", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return JobChangeEvent


class JobCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Job", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return JobChangeEvent


class JobDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Job", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return JobChangeEvent


class JobUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Job", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return JobChangeEvent


# Namespace Triggers
class NamespaceAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Namespace", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NamespaceChangeEvent


class NamespaceCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Namespace", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NamespaceChangeEvent


class NamespaceDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Namespace", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NamespaceChangeEvent


class NamespaceUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Namespace", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return NamespaceChangeEvent


# ServiceAccount Triggers
class ServiceAccountAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ServiceAccount", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceAccountChangeEvent


class ServiceAccountCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ServiceAccount", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceAccountChangeEvent


class ServiceAccountDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ServiceAccount", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceAccountChangeEvent


class ServiceAccountUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="ServiceAccount", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return ServiceAccountChangeEvent


# PersistentVolume Triggers
class PersistentVolumeAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="PersistentVolume", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PersistentVolumeChangeEvent


class PersistentVolumeCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="PersistentVolume", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PersistentVolumeChangeEvent


class PersistentVolumeDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="PersistentVolume", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PersistentVolumeChangeEvent


class PersistentVolumeUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="PersistentVolume", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return PersistentVolumeChangeEvent


# Kubernetes Any Triggers
class KubernetesAnyAllChangesTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Any", 
            operation=None, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return KubernetesAnyChangeEvent


class KubernetesAnyCreateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Any", 
            operation=K8sOperationType.CREATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return KubernetesAnyChangeEvent


class KubernetesAnyDeleteTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Any", 
            operation=K8sOperationType.DELETE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return KubernetesAnyChangeEvent


class KubernetesAnyUpdateTrigger(K8sBaseTrigger):

    def __init__(self, name_prefix: str = None, namespace_prefix: str = None, labels_selector: str = None):
        super().__init__(
            kind="Any", 
            operation=K8sOperationType.UPDATE, 
            name_prefix=name_prefix, 
            namespace_prefix=namespace_prefix,
            labels_selector=labels_selector,
        )

    @staticmethod
    def get_execution_event_type() -> type:
        return KubernetesAnyChangeEvent


# K8s Trigger class
class K8sTriggers(BaseModel):

    on_clusterrole_all_changes: Optional[ClusterRoleAllChangesTrigger]
    on_clusterrole_create: Optional[ClusterRoleCreateTrigger]
    on_clusterrole_delete: Optional[ClusterRoleDeleteTrigger]
    on_clusterrole_update: Optional[ClusterRoleUpdateTrigger]
    on_clusterrolebinding_all_changes: Optional[ClusterRoleBindingAllChangesTrigger]
    on_clusterrolebinding_create: Optional[ClusterRoleBindingCreateTrigger]
    on_clusterrolebinding_delete: Optional[ClusterRoleBindingDeleteTrigger]
    on_clusterrolebinding_update: Optional[ClusterRoleBindingUpdateTrigger]
    on_daemonset_all_changes: Optional[DaemonSetAllChangesTrigger]
    on_daemonset_create: Optional[DaemonSetCreateTrigger]
    on_daemonset_delete: Optional[DaemonSetDeleteTrigger]
    on_daemonset_update: Optional[DaemonSetUpdateTrigger]
    on_deployment_all_changes: Optional[DeploymentAllChangesTrigger]
    on_deployment_create: Optional[DeploymentCreateTrigger]
    on_deployment_delete: Optional[DeploymentDeleteTrigger]
    on_deployment_update: Optional[DeploymentUpdateTrigger]
    on_event_all_changes: Optional[EventAllChangesTrigger]
    on_event_create: Optional[EventCreateTrigger]
    on_event_delete: Optional[EventDeleteTrigger]
    on_event_update: Optional[EventUpdateTrigger]
    on_horizontalpodautoscaler_all_changes: Optional[HorizontalPodAutoscalerAllChangesTrigger]
    on_horizontalpodautoscaler_create: Optional[HorizontalPodAutoscalerCreateTrigger]
    on_horizontalpodautoscaler_delete: Optional[HorizontalPodAutoscalerDeleteTrigger]
    on_horizontalpodautoscaler_update: Optional[HorizontalPodAutoscalerUpdateTrigger]
    on_job_all_changes: Optional[JobAllChangesTrigger]
    on_job_create: Optional[JobCreateTrigger]
    on_job_delete: Optional[JobDeleteTrigger]
    on_job_update: Optional[JobUpdateTrigger]
    on_kubernetes_any_resource_all_changes: Optional[KubernetesAnyAllChangesTrigger]
    on_kubernetes_any_resource_create: Optional[KubernetesAnyCreateTrigger]
    on_kubernetes_any_resource_delete: Optional[KubernetesAnyDeleteTrigger]
    on_kubernetes_any_resource_update: Optional[KubernetesAnyUpdateTrigger]
    on_namespace_all_changes: Optional[NamespaceAllChangesTrigger]
    on_namespace_create: Optional[NamespaceCreateTrigger]
    on_namespace_delete: Optional[NamespaceDeleteTrigger]
    on_namespace_update: Optional[NamespaceUpdateTrigger]
    on_node_all_changes: Optional[NodeAllChangesTrigger]
    on_node_create: Optional[NodeCreateTrigger]
    on_node_delete: Optional[NodeDeleteTrigger]
    on_node_update: Optional[NodeUpdateTrigger]
    on_persistentvolume_all_changes: Optional[PersistentVolumeAllChangesTrigger]
    on_persistentvolume_create: Optional[PersistentVolumeCreateTrigger]
    on_persistentvolume_delete: Optional[PersistentVolumeDeleteTrigger]
    on_persistentvolume_update: Optional[PersistentVolumeUpdateTrigger]
    on_pod_all_changes: Optional[PodAllChangesTrigger]
    on_pod_create: Optional[PodCreateTrigger]
    on_pod_delete: Optional[PodDeleteTrigger]
    on_pod_update: Optional[PodUpdateTrigger]
    on_replicaset_all_changes: Optional[ReplicaSetAllChangesTrigger]
    on_replicaset_create: Optional[ReplicaSetCreateTrigger]
    on_replicaset_delete: Optional[ReplicaSetDeleteTrigger]
    on_replicaset_update: Optional[ReplicaSetUpdateTrigger]
    on_service_all_changes: Optional[ServiceAllChangesTrigger]
    on_service_create: Optional[ServiceCreateTrigger]
    on_service_delete: Optional[ServiceDeleteTrigger]
    on_service_update: Optional[ServiceUpdateTrigger]
    on_serviceaccount_all_changes: Optional[ServiceAccountAllChangesTrigger]
    on_serviceaccount_create: Optional[ServiceAccountCreateTrigger]
    on_serviceaccount_delete: Optional[ServiceAccountDeleteTrigger]
    on_serviceaccount_update: Optional[ServiceAccountUpdateTrigger]
    on_statefulset_all_changes: Optional[StatefulSetAllChangesTrigger]
    on_statefulset_create: Optional[StatefulSetCreateTrigger]
    on_statefulset_delete: Optional[StatefulSetDeleteTrigger]
    on_statefulset_update: Optional[StatefulSetUpdateTrigger]



import logging
import os
from collections import defaultdict
from enum import Enum
from time import sleep
import sentry_sdk
import requests
import threading
from ..model.config import Registry, Telemetry
from .telemetry import SinkInfo

from hikaru.model import NodeList

class TelemetryLevel(Enum):
    NONE = 0,
    USAGE = 1,
    ERROR = 2 

class TelemetryService: 
    def __init__(
        self,
        telemetry_level: TelemetryLevel,
        endpoint: str,
        periodic_time_sec: float,
        registry : Registry
    ):
        self.telemetry_level = telemetry_level
        self.endpoint = endpoint
        self.registry = registry
        self.periodic_time_sec = periodic_time_sec

        sentry_dsn = os.environ.get("SENTRY_DSN", "")
        if self.telemetry_level == TelemetryLevel.ERROR and sentry_dsn:
            logging.info(f"Telemetry set to include error info, Thank you for helping us improve Robusta.")
            try:
                sentry_sdk.init(sentry_dsn, traces_sample_rate=float(os.environ.get("SENTRY_TRACES_SAMPLE_RATE", 0.5)))
            except Exception as e:
                logging.error(f"Sentry error: {e}", exc_info=True)

        self.__thread = threading.Thread(target=self.__log_periodic)
        self.__thread.start()


    def __log_periodic(self):
        while(True):
            try:
                tele = self.registry.get_telemetry()
            
                current_nodes: NodeList = NodeList.listNode().obj
                tele.nodes_count = len(current_nodes.items)

                self.__log(tele)

                tele.sinks_info = defaultdict(lambda: SinkInfo())

                sleep(self.periodic_time_sec)
            except Exception as e:
                logging.error(f"Failed to run periodic telemetry update {e}", exc_info=True)
                pass


    def __log(self, data: Telemetry):
        r = requests.post(self.endpoint, data=data.json(), headers={'Content-Type': 'application/json'})
        if(r.status_code != 201):
            logging.error(f"Failed to log telemetry data")

        return r



import logging
import opsgenie_sdk
from typing import List

from .opsgenie_sink_params import OpsGenieSinkConfigWrapper
from ..transformer import Transformer
from ...reporting.base import (
    Finding,
    Enrichment,
    FindingSeverity,
)
from ..sink_base import SinkBase

# Robusta has only 4 severities, OpsGenie has 5. We had to omit 1 priority
PRIORITY_MAP = {
    FindingSeverity.INFO: "P5",
    FindingSeverity.LOW: "P4",
    FindingSeverity.MEDIUM: "P3",
    FindingSeverity.HIGH: "P1"
}


class OpsGenieSink(SinkBase):
    def __init__(self, sink_config: OpsGenieSinkConfigWrapper, registry):
        super().__init__(sink_config.opsgenie_sink, registry)

        self.api_key = sink_config.opsgenie_sink.api_key
        self.teams = sink_config.opsgenie_sink.teams
        self.tags = sink_config.opsgenie_sink.tags

        self.conf = opsgenie_sdk.configuration.Configuration()
        self.conf.api_key['Authorization'] = self.api_key

        self.api_client = opsgenie_sdk.api_client.ApiClient(configuration=self.conf)
        self.alert_api = opsgenie_sdk.AlertApi(api_client=self.api_client)

    def write_finding(self, finding: Finding, platform_enabled: bool):
        description = self.__to_description(finding, platform_enabled)
        self.tags.insert(0, self.cluster_name)
        body = opsgenie_sdk.CreateAlertPayload(
            source="Robusta",
            message=finding.title,
            description=description,
            alias=finding.title,
            responders=[
                {
                    "name": team,
                    "type": "team"
                } for team in self.teams
            ],
            tags=self.tags,
            entity=finding.service_key,
            priority=PRIORITY_MAP.get(finding.severity, "P3")
        )
        try:
            self.alert_api.create_alert(create_alert_payload=body)
        except opsgenie_sdk.ApiException as err:
            logging.error(f"Error creating opsGenie alert {finding} {err}", exc_info=True)

    def __to_description(self, finding: Finding, platform_enabled: bool) -> str:
        description = ""
        if platform_enabled:
            description = f"<a href=\"{finding.investigate_uri}\">Investigate</a>\n"

        return f"{description}{self.__enrichments_as_text(finding.enrichments)}"

    @classmethod
    def __enrichments_as_text(cls, enrichments: List[Enrichment]) -> str:
        text_arr = [
            Transformer.to_html(enrichment.blocks) for enrichment in enrichments
        ]
        return "---\n".join(text_arr)




import logging
import time

import requests


class WebApi:
    @staticmethod
    def run_manual_action(action_name: str, action_params=None, sinks=None, retries=1, timeout_delay=1):
        if sinks is None:
            sinks = []
        if action_params is None:
            action_params = {"annotation": None}
        status_code = -1

        manual_action_url = "http://127.0.0.1:5000/api/trigger"
        data = { "action_name": action_name, "action_params": action_params, "sinks": sinks }
        for _ in range(retries):
            try:
                response = requests.post(
                    manual_action_url,
                    json=data
                )
                status_code = response.status_code
                if status_code == 200:
                    return status_code

                logging.error(
                    f"Failed to run manual action \naction_name:{action_name}\n"
                    f"Reason: {response.reason}\nStatus Code{status_code}"
                )
                time.sleep(timeout_delay)
            except Exception as e:
                logging.error(f"Error sending manual action request {e}")
                time.sleep(timeout_delay)

        return status_code



import json
import logging
import threading
import time
import traceback
from typing import List, Dict, Any
from supabase_py.lib.auth_client import SupabaseAuthClient

from .model_conversion import ModelConversion
from ....model.cluster_status import ClusterStatus
from ....model.nodes import NodeInfo
from ....model.services import ServiceInfo
from ....reporting.base import Finding
from ....model.env_vars import SUPABASE_LOGIN_RATE_LIMIT_SEC
from supabase_py import Client

SERVICES_TABLE = "Services"
NODES_TABLE = "Nodes"
EVIDENCE_TABLE = "Evidence"
ISSUES_TABLE = "Issues"
CLUSTERS_STATUS_TABLE = "ClustersStatus"


class RobustaAuthClient(SupabaseAuthClient):
    def _set_timeout(*args, **kwargs):
        """Set timer task"""
        # _set_timeout isn't implemented in gotrue client. it's required for the jwt refresh token timer task
        # https://github.com/supabase/gotrue-py/blob/49c092e3a4a6d7bb5e1c08067a4c42cc2f74b5cc/gotrue/client.py#L242
        # callback, timeout_ms
        threading.Timer(args[2] / 1000, args[1]).start()


class RobustaClient(Client):
    def _get_auth_headers(self) -> Dict[str, str]:
        auth = getattr(self, "auth", None)
        session = auth.current_session if auth else None
        if session and session["access_token"]:
            access_token = auth.session()["access_token"]
        else:
            access_token = self.supabase_key

        headers: Dict[str, str] = {
            "apiKey": self.supabase_key,
            "Authorization": f"Bearer {access_token}",
        }
        return headers

    @staticmethod
    def _init_supabase_auth_client(
        auth_url: str,
        supabase_key: str,
        detect_session_in_url: bool,
        auto_refresh_token: bool,
        persist_session: bool,
        local_storage: Dict[str, Any],
        headers: Dict[str, str],
    ) -> RobustaAuthClient:
        """Creates a wrapped instance of the GoTrue Client."""
        return RobustaAuthClient(
            url=auth_url,
            auto_refresh_token=auto_refresh_token,
            detect_session_in_url=detect_session_in_url,
            persist_session=persist_session,
            local_storage=local_storage,
            headers=headers,
        )


class SupabaseDal:
    def __init__(
        self,
        url: str,
        key: str,
        account_id: str,
        email: str,
        password: str,
        sink_name: str,
        cluster_name: str,
        signing_key: str,
    ):
        self.url = url
        self.key = key
        self.account_id = account_id
        self.cluster = cluster_name
        self.client = RobustaClient(url, key)
        self.email = email
        self.password = password
        self.sign_in_time = 0
        self.sign_in()
        self.sink_name = sink_name
        self.signing_key = signing_key

    def persist_finding(self, finding: Finding):
        for enrichment in finding.enrichments:
            res = (
                self.client.table(EVIDENCE_TABLE)
                .insert(ModelConversion.to_evidence_json(
                    account_id=self.account_id,
                    cluster_id=self.cluster,
                    sink_name=self.sink_name,
                    signing_key=self.signing_key,
                    finding_id=finding.id,
                    enrichment=enrichment
                ))
                .execute()
            )
            if res.get("status_code") != 201:
                logging.error(
                    f"Failed to persist finding {finding.id} enrichment {enrichment} error: {res.get('data')}"
                )

        res = self.client.table(ISSUES_TABLE).insert(
            ModelConversion.to_finding_json(self.account_id, self.cluster, finding)
        ).execute()
        if res.get("status_code") != 201:
            logging.error(
                f"Failed to persist finding {finding.id} error: {res.get('data')}"
            )
            self.handle_supabase_error()

    def to_service(self, service: ServiceInfo) -> Dict[Any, Any]:
        return {
            "name": service.name,
            "type": service.service_type,
            "namespace": service.namespace,
            "classification": service.classification,
            "cluster": self.cluster,
            "account_id": self.account_id,
            "deleted": service.deleted,
            "service_key": service.get_service_key(),
            "update_time": "now()",
        }

    def persist_service(self, service: ServiceInfo):
        db_service = self.to_service(service)
        res = (
            self.client.table(SERVICES_TABLE).insert(db_service, upsert=True).execute()
        )
        if res.get("status_code") not in [200, 201]:
            logging.error(
                f"Failed to persist service {service} error: {res.get('data')}"
            )
            self.handle_supabase_error()

    def get_active_services(self) -> List[ServiceInfo]:
        res = (
            self.client.table(SERVICES_TABLE)
            .select("name", "type", "namespace", "classification")
            .filter("account_id", "eq", self.account_id)
            .filter("cluster", "eq", self.cluster)
            .filter("deleted", "eq", False)
            .execute()
        )
        if res.get("status_code") not in [200]:
            msg = f"Failed to get existing services (supabase) error: {res.get('data')}"
            logging.error(msg)
            self.handle_supabase_error()
            raise Exception(msg)

        return [
            ServiceInfo(
                name=service["name"],
                service_type=service["type"],
                namespace=service["namespace"],
                classification=service["classification"],
            )
            for service in res.get("data")
        ]

    def has_cluster_findings(self) -> bool:
        res = (
            self.client.table(ISSUES_TABLE)
            .select('*')
            .filter("account_id", "eq", self.account_id)
            .filter("cluster", "eq", self.cluster)
            .limit(1)
            .execute()
        )
        if res.get("status_code") not in [200]:
            msg = f"Failed to check cluster issues: {res.get('data')}"
            logging.error(msg)
            self.handle_supabase_error()
            raise Exception(msg)

        return len(res.get("data")) > 0

    def get_active_nodes(self) -> List[NodeInfo]:
        res = (
            self.client.table(NODES_TABLE)
            .select("*")
            .filter("account_id", "eq", self.account_id)
            .filter("cluster_id", "eq", self.cluster)
            .filter("deleted", "eq", False)
            .execute()
        )
        if res.get("status_code") not in [200]:
            msg = f"Failed to get existing nodes (supabase) error: {res.get('data')}"
            logging.error(msg)
            self.handle_supabase_error()
            raise Exception(msg)

        return [
            NodeInfo(
                name=node["name"],
                node_creation_time=node["node_creation_time"],
                taints=node["taints"],
                conditions=node["conditions"],
                memory_capacity=node["memory_capacity"],
                memory_allocatable=node["memory_allocatable"],
                memory_allocated=node["memory_allocated"],
                cpu_capacity=node["cpu_capacity"],
                cpu_allocatable=node["cpu_allocatable"],
                cpu_allocated=node["cpu_allocated"],
                pods_count=node["pods_count"],
                pods=node["pods"],
                internal_ip=node["internal_ip"],
                external_ip=node["external_ip"],
                node_info=json.loads(node["node_info"]),
            )
            for node in res.get("data")
        ]

    def __to_db_node(self, node: NodeInfo) -> Dict[Any, Any]:
        db_node = node.dict()
        db_node["account_id"] = self.account_id
        db_node["cluster_id"] = self.cluster
        db_node["updated_at"] = "now()"
        return db_node

    def publish_node(self, node: NodeInfo):
        db_node = self.__to_db_node(node)
        res = (
            self.client.table(NODES_TABLE).insert(db_node, upsert=True).execute()
        )
        if res.get("status_code") not in [200, 201]:
            logging.error(
                f"Failed to persist node {node} error: {res.get('data')}"
            )
            self.handle_supabase_error()

    def sign_in(self):
        if time.time() > self.sign_in_time + SUPABASE_LOGIN_RATE_LIMIT_SEC:
            logging.info("Supabase dal login")
            self.sign_in_time = time.time()
            self.client.auth.sign_in(email=self.email, password=self.password)

    def handle_supabase_error(self):
        """Workaround for Gotrue bug in refresh token."""
        # If there's an error during refresh token, no new refresh timer task is created, and the client remains not authenticated for good
        # When there's an error connecting to supabase server, we will re-login, to re-authenticate the session.
        # Adding rate-limiting mechanism, not to login too much because of other errors
        # https://github.com/supabase/gotrue-py/issues/9
        try:
            self.sign_in()
        except Exception as e:
            logging.error(f"Failed to sign in on error", exc_info=True)

    def to_db_cluster_status(self, data: ClusterStatus) -> Dict[str, Any]:
        db_cluster_status = data.dict()
        if data.last_alert_at is  None:
            del db_cluster_status["last_alert_at"]
            
        db_cluster_status["updated_at"] = "now()"    
        logging.info(f"cluster status {db_cluster_status}")        
        return db_cluster_status

    def publish_cluster_status(self, cluster_status: ClusterStatus):
        res = (
            self.client.table(CLUSTERS_STATUS_TABLE).insert(self.to_db_cluster_status(cluster_status), upsert=True).execute()
        )
        if res.get("status_code") not in [200, 201]:
            logging.error(
                f"Failed to upsert {self.to_db_cluster_status(cluster_status)} error: {res.get('data')}"
            )
            self.handle_supabase_error()



from typing import List

from .msteams_base import MsTeamsBase
from .msteams_action import MsTeamsAction


class MsTeamsColumn(MsTeamsBase):
    def __init__(self) -> None:
        self.column_list = []
        super().__init__(self.__column_set())

    def __column_set(self) -> map:
        return {
            "type": "ColumnSet",
            "columns": self.column_list
        }

    def add_column(
        self,
        items: List[MsTeamsBase],
        width_stretch: bool = False,
        is_visible: bool = True,
        key: str = None,
        action: MsTeamsAction = None,
    ):
        block = {
            "type": "Column",
            "width": "stretch" if width_stretch else "auto",
            "isVisible": is_visible,
            "items": self.__to_map_list(items),
        }

        if key is not None:
            block["id"] = key
        if action is not None:
            block.update(action.get_map_value())

        self.column_list.append(block)

    def __to_map_list(self, elements: List[MsTeamsBase]) -> List[dict]:
        return [element.get_map_value() for element in elements]



from typing import List

from .msteams_column import MsTeamsColumn
from .msteams_text_block import MsTeamsTextBlock
from .msteams_base import MsTeamsBase


class MsTeamsTable(MsTeamsBase):
    def __init__(self, headers: List[str], rows: List[List[str]]):

        column_element = self.__create_table(headers, rows)
        super().__init__(column_element.get_map_value())

    def __create_table(self, headers: List[str], rows: List[List[str]]) -> MsTeamsColumn:

        column_element = MsTeamsColumn()
        for index in range(len(headers)):
            single_column = [MsTeamsTextBlock(text=headers[index], weight="bolder")]
            single_column = single_column + self.__create_single_column_list(rows=rows, index=index)
            column_element.add_column(items=single_column, width_stretch=True)

        return column_element

    def __create_single_column_list(self, rows: List[List[str]], index: int) -> List[map]:
        first_row = True
        column = []
        for row in rows:
            column.append(MsTeamsTextBlock(text=row[index], separator=first_row))
            first_row = False
        return column



from ..sink_config import SinkConfigBase
from ..sink_base_params import SinkBaseParams


class TelegramSinkParams(SinkBaseParams):
    bot_token: str
    chat_id: int
    send_files: bool = True  # Change to False, to omit file attachments


class TelegramSinkConfigWrapper(SinkConfigBase):
    telegram_sink: TelegramSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.telegram_sink




from typing import Optional
from hikaru.model import EventList, Event

from ..reporting import TableBlock
from ..reporting.custom_rendering import RendererType
from ...integrations.kubernetes.api_client_utils import parse_kubernetes_datetime_to_ms


def get_resource_events_table(table_name: str, kind: str, name: str, namespace: str = None) -> Optional[TableBlock]:
    field_selector = f"involvedObject.kind={kind},involvedObject.name={name}"
    if namespace:
        field_selector += f",involvedObject.namespace={namespace}"

    event_list: EventList = Event.listEventForAllNamespaces(field_selector=field_selector).obj

    if event_list.items:
        headers = ["reason", "type", "time", "message"]
        rows = [
            [event.reason,
             event.type,
             parse_kubernetes_datetime_to_ms(get_event_timestamp(event)) if get_event_timestamp(event) else 0,
             event.message]
            for event in event_list.items
        ]
        return TableBlock(
                rows=rows,
                headers=headers,
                column_renderers={"time": RendererType.DATETIME},
                table_name=table_name
            )
    return None


def get_event_timestamp(event: Event):
    if event.lastTimestamp:
        return event.lastTimestamp
    elif event.eventTime:
        return event.eventTime
    elif event.firstTimestamp:
        return event.firstTimestamp
    if event.metadata.creationTimestamp:
        return event.metadata.creationTimestamp
    return


def get_events_list(event_type: str = None) -> EventList:
    """
    event_types are ["Normal","Warning"]
    """
    if event_type:
        field_selector = f"type={event_type}"
        return Event.listEventForAllNamespaces(field_selector=field_selector).obj
    return Event.listEventForAllNamespaces().obj





import re
class MsTeamsMarkDownFixUrl:

    @staticmethod
    def fix_text(text: str):
        
        pattern = r'<http.*\|.*>'

        new_text = text
        for bad_url in re.findall(pattern, text):
            good_url = MsTeamsMarkDownFixUrl.__fix_url(bad_url)
            new_text = new_text.replace(bad_url, good_url)
            
        return new_text

    @staticmethod
    def __fix_url(bad_url : str):
        parts = bad_url.replace('<', '').replace('>','').split('|')
        url = '(' + parts[0] + ')'
        description = '[' + parts[1] + ']'

        return description + url






from typing import List
from xml.etree import ElementTree

from docutils.core import publish_doctree


class DocstringField:
    """
    Represents a docstring field like ":var xyz: abc"
    For the above example, field_type is "var", field_target is "xyz", and field_value is "abc"
    """

    def __init__(self, name: str, body: str):
        field_type, field_target = name.split()
        self.field_type = field_type
        self.field_target = field_target
        self.field_value = body


class Docstring:
    def __init__(self, docstring: str):
        """
        Parse docutils/sphinx-style docs from a docstring.

        For example, given the following docstring:

            '''
            This is the description

            :var x: abc
            :example x: efg
            '''

        We parse:
        - description="This is the description"
        - fields:
            - var field
            - example field

        :param docstring: the docstring to parse
        """
        dom = publish_doctree(docstring).asdom()
        tree = ElementTree.fromstring(dom.toxml())
        self.fields = []

        for field in tree.iter(tag="field"):
            name = next(field.iter(tag="field_name"))
            body = next(field.iter(tag="field_body"))
            self.fields.append(DocstringField(name.text, "".join(body.itertext())))

        # we select all the docutils paragraphs which are under a root <block_quote> element
        # this is essentially all the lines in the docstring which don't contain :field:
        self.description = "\n\n".join(
            ["".join(p.itertext()) for p in tree.findall("./block_quote/paragraph")]
        )



import os
import uuid
import pytz

PLAYBOOKS_ROOT = os.environ.get("PLAYBOOKS_ROOT", "/etc/robusta/playbooks/")
# should be the same as the one in out Dockerfile
DEFAULT_PLAYBOOKS_ROOT = os.environ.get("DEFAULT_PLAYBOOKS_ROOT", os.path.join(PLAYBOOKS_ROOT, "defaults"))
CUSTOM_PLAYBOOKS_ROOT = os.path.join(PLAYBOOKS_ROOT, "storage")

PLAYBOOKS_CONFIG_FILE_PATH = os.environ.get("PLAYBOOKS_CONFIG_FILE_PATH")

INSTALLATION_NAMESPACE = os.environ.get("INSTALLATION_NAMESPACE", "robusta")
DISCOVERY_PERIOD_SEC = int(os.environ.get("DISCOVERY_PERIOD_SEC", 90))
PERIODIC_LONG_SEC = int(os.environ.get("PERIODIC_LONG_SEC", 60 * 60 * 3))
SUPABASE_LOGIN_RATE_LIMIT_SEC = int(
    os.environ.get("SUPABASE_LOGIN_RATE_LIMIT_SEC", 900)
)
GRAFANA_RENDERER_URL = os.environ.get(
    "GRAFANA_RENDERER_URL", "http://127.0.0.1:8281/render"
)
SERVICE_UPDATES_CACHE_TTL_SEC = os.environ.get("SERVICE_UPDATES_CACHE_TTL_SEC", 120)
INTERNAL_PLAYBOOKS_ROOT = os.environ.get(
    "INTERNAL_PLAYBOOKS_ROOT", "/app/src/robusta/core/playbooks/internal"
)
DEFAULT_TIMEZONE = pytz.timezone(os.environ.get("DEFAULT_TIMEZONE", "UTC"))
NUM_EVENT_THREADS = int(os.environ.get("NUM_EVENT_THREADS", 10))
INCOMING_EVENTS_QUEUE_MAX_SIZE = int(
    os.environ.get("INCOMING_EVENTS_QUEUE_MAX_SIZE", 500)
)

FLOAT_PRECISION_LIMIT = int(os.environ.get("FLOAT_PRECISION_LIMIT", 11))

PROMETHEUS_REQUEST_TIMEOUT_SECONDS = float(
    os.environ.get("PROMETHEUS_REQUEST_TIMEOUT_SECONDS", 90.0)
)
PROMETHEUS_ENABLED = os.environ.get("PROMETHEUS_ENABLED", "false").lower() == "true"

INCOMING_REQUEST_TIME_WINDOW_SECONDS = int(
    os.environ.get("INCOMING_REQUEST_TIME_WINDOW_SECONDS", 3600)
)

RELAY_EXTERNAL_ACTIONS_URL = os.environ.get(
    "RELAY_EXTERNAL_ACTIONS_URL", "https://api.robusta.dev/integrations/generic/actions"
)

GIT_MAX_RETRIES = int(os.environ.get("GIT_MAX_RETRIES", 100))

PRINTED_TABLE_MAX_WIDTH = int(os.environ.get("PRINTED_TABLE_MAX_WIDTH", 70))

RUNNER_VERSION = os.environ.get("RUNNER_VERSION", "unknown")

GRAFANA_READ_TIMEOUT = int(os.environ.get("GRAFANA_READ_TIMEOUT", 20))

TEAMS_IMAGE_WIDTH = os.environ.get("TEAMS_IMAGE_WIDTH", "700px")

ROBUSTA_UI_DOMAIN = os.environ.get("ROBUSTA_UI_DOMAIN", "https://platform.robusta.dev")

ROBUSTA_TELEMETRY_ENDPOINT = os.environ.get("ROBUSTA_TELEMETRY_ENDPOINT", "https://api.robusta.dev/telemetry")
ENABLE_TELEMETRY = os.environ.get("ENABLE_TELEMETRY", "true").lower() == "true"
SEND_ADDITIONAL_TELEMETRY = os.environ.get("SEND_ADDITIONAL_TELEMETRY", "false").lower() == "true"
TELEMETRY_PERIODIC_SEC = int(os.environ.get("TELEMETRY_PERIODIC_SEC", 60 * 60 * 24)) # 24H

SLACK_TABLE_COLUMNS_LIMIT = int(os.environ.get("SLACK_TABLE_COLUMNS_LIMIT", 4))
RSA_KEYS_PATH = os.environ.get("RSA_KEYS_PATH", "/etc/robusta/auth")



import base64
import json
import random
import subprocess
import time
import urllib.request
import uuid
import click_spinner
from distutils.version import StrictVersion
from typing import Optional, List, Union, Dict
from zipfile import ZipFile
import traceback

import requests
import typer
import yaml
from kubernetes import config
from pydantic import BaseModel, Extra

# TODO - separate shared classes to a separated shared repo, to remove dependencies between the cli and runner
from .auth import gen_rsa_pair, RSAKeyPair
from .backend_profile import backend_profile
from ..core.sinks.msteams.msteams_sink_params import (
    MsTeamsSinkConfigWrapper,
    MsTeamsSinkParams,
)
from ..core.sinks.robusta.robusta_sink_params import (
    RobustaSinkConfigWrapper,
    RobustaSinkParams,
)
from ..core.sinks.slack.slack_sink_params import SlackSinkConfigWrapper, SlackSinkParams
from robusta._version import __version__
from .integrations_cmd import app as integrations_commands, get_slack_key
from .auth import app as auth_commands
from .slack_verification import verify_slack_channel
from .slack_feedback_message import SlackFeedbackMessagesSender, SlackFeedbackConfig
from .playbooks_cmd import app as playbooks_commands
from .utils import log_title, replace_in_file, namespace_to_kubectl

FORWARDER_CONFIG_FOR_SMALL_CLUSTERS = "64Mi"
RUNNER_CONFIG_FOR_SMALL_CLUSTERS = "512Mi"
GRAFANA_RENDERER_CONFIG_FOR_SMALL_CLUSTERS = "64Mi"

app = typer.Typer()
app.add_typer(playbooks_commands, name="playbooks", help="Playbooks commands menu")
app.add_typer(
    integrations_commands, name="integrations", help="Integrations commands menu"
)
app.add_typer(
    auth_commands, name="auth", help="Authentication commands menu"
)


def get_runner_url(runner_version=None):
    if runner_version is None:
        runner_version = __version__
    return f"https://gist.githubusercontent.com/robusta-lab/6b809d508dfc3d8d92afc92c7bbbe88e/raw/robusta-{runner_version}.yaml"


class GlobalConfig(BaseModel):
    signing_key: str = ""
    account_id: str = ""


class PodConfigs(Dict[str, Dict[str, Dict[str, str]]]):
    __root__: Dict[str, Dict[str, Dict[str, str]]]

    @classmethod
    def gen_config(cls, memory_size: str) -> Dict:
        return {"resources": {"requests": {"memory": memory_size}}}


class HelmValues(BaseModel, extra=Extra.allow):
    globalConfig: GlobalConfig
    sinksConfig: List[
        Union[
            SlackSinkConfigWrapper, RobustaSinkConfigWrapper, MsTeamsSinkConfigWrapper
        ]
    ]
    clusterName: str
    enablePrometheusStack: bool = False
    disableCloudRouting: bool = False
    enablePlatformPlaybooks: bool = False
    playbooksPersistentVolumeSize: str = None
    kubewatch: Dict = None
    grafanaRenderer: Dict = None
    runner: Dict = None
    rsa: RSAKeyPair = None

    def set_pod_configs_for_small_clusters(self):
        self.kubewatch = PodConfigs.gen_config(FORWARDER_CONFIG_FOR_SMALL_CLUSTERS)
        self.runner = PodConfigs.gen_config(RUNNER_CONFIG_FOR_SMALL_CLUSTERS)
        self.grafanaRenderer = PodConfigs.gen_config(
            GRAFANA_RENDERER_CONFIG_FOR_SMALL_CLUSTERS
        )


def guess_cluster_name(context):
    with click_spinner.spinner():
        try:
            all_contexts, current_context = config.list_kube_config_contexts()
            if context is not None:
                for i in range(len(all_contexts)):
                    if all_contexts[i].get('name') == context:
                        return all_contexts[i].get('context').get('cluster')
                typer.echo(f" no context exists with the name '{context}', your current context is {current_context.get('cluster')}")
            if current_context and current_context.get("name"):
                return current_context.get("context").get("cluster")
        except Exception:  # this happens, for example, if you don't have a kubeconfig file
            typer.echo("Error reading kubeconfig to generate cluster name")
        return f"cluster_{random.randint(0, 1000000)}"


def get_slack_channel() -> str:
    return (
        typer.prompt(
            "Which slack channel should I send notifications to? ",
            prompt_suffix="#",
        )
        .strip()
        .strip("#")
    )


def write_values_file(output_path: str, values: HelmValues):
    with open(output_path, "w") as output_file:
        yaml.safe_dump(values.dict(exclude_defaults=True), output_file, sort_keys=False)
        typer.secho(
            f"Saved configuration to {output_path}",
            fg="green",
        )
        typer.secho(
            f"Save this file for future use. It contains your account credentials",
            fg="red",
        )


@app.command()
def gen_config(
    cluster_name: str = typer.Option(
        None,
        help="Cluster Name",
    ),
    is_small_cluster: bool = typer.Option(None),
    slack_api_key: str = typer.Option(
        "",
        help="Slack API Key",
    ),
    slack_channel: str = typer.Option(
        "",
        help="Slack Channel",
    ),
    msteams_webhook: str = typer.Option(
        None,
        help="MsTeams webhook url",
    ),
    robusta_api_key: str = typer.Option(None),
    enable_prometheus_stack: bool = typer.Option(None),
    disable_cloud_routing: bool = typer.Option(None),
    output_path: str = typer.Option(
        "./generated_values.yaml", help="Output path of generated Helm values"
    ),
    debug: bool = typer.Option(False),
    context: str = typer.Option(
        None,
        help="The name of the kubeconfig context to use",
    ),
):
    """Create runtime configuration file"""
    if cluster_name is None:
        cluster_name = typer.prompt(
            "Please specify a unique name for your cluster or press ENTER to use the default",
            default=guess_cluster_name(context),
        )
    if is_small_cluster is None:
        is_small_cluster = typer.confirm(
            "Are you running a mini cluster? (Like minikube or kind)"
        )

    sinks_config: List[
        Union[
            SlackSinkConfigWrapper, RobustaSinkConfigWrapper, MsTeamsSinkConfigWrapper
        ]
    ] = []
    slack_workspace = "N/A"
    if not slack_api_key and typer.confirm(
        "Do you want to configure slack integration? This is HIGHLY recommended.",
        default=True,
    ):
        slack_api_key, slack_workspace = get_slack_key()

    if slack_api_key and not slack_channel:
        slack_channel = get_slack_channel()

    slack_integration_configured = False
    if slack_api_key and slack_channel:
        while not verify_slack_channel(
            slack_api_key, cluster_name, slack_channel, slack_workspace, debug
        ):
            slack_channel = get_slack_channel()

        sinks_config.append(
            SlackSinkConfigWrapper(
                slack_sink=SlackSinkParams(
                    name="main_slack_sink",
                    api_key=slack_api_key,
                    slack_channel=slack_channel,
                )
            )
        )

        slack_integration_configured = True

    if msteams_webhook is None and typer.confirm(
        "Do you want to configure MsTeams integration ?",
        default=False,
    ):
        msteams_webhook = typer.prompt(
            "Please insert your MsTeams webhook url",
            default=None,
        )

    if msteams_webhook:
        sinks_config.append(
            MsTeamsSinkConfigWrapper(
                ms_teams_sink=MsTeamsSinkParams(
                    name="main_ms_teams_sink",
                    webhook_url=msteams_webhook,
                )
            )
        )

    enable_platform_playbooks = False
    # we have a slightly different flow here than the other options so that pytest can pass robusta_api_key="" to skip
    # asking the question
    if robusta_api_key is None:
        if typer.confirm(
            "Would you like to use Robusta UI? This is HIGHLY recommended.",
            default=True,
        ):
            if typer.confirm("Do you already have a Robusta account?"):
                while True:
                    robusta_api_key = typer.prompt(
                        "Please insert your Robusta account token",
                        default=None,
                    )
                    try:
                        json.loads(base64.b64decode(robusta_api_key))
                        break
                    except Exception:
                        typer.secho(
                            "Sorry, invalid token format. "
                            "The token can be found in any existing generated_values.yaml file, under the robusta_sink",
                            fg="red",
                        )

            else:  # self registration
                account_name = typer.prompt("Choose your account name")
                email = typer.prompt(
                    "Enter a Gmail/Google Workspace address. This will be used to login"
                )
                email = email.strip()
                res = requests.post(
                    f"{backend_profile.robusta_cloud_api_host}/accounts/create",
                    json={
                        "account_name": account_name,
                        "email": email,
                    },
                )
                if res.status_code == 201:
                    robusta_api_key = res.json().get("token")
                    typer.echo(
                        "Successfully registered.\n",
                        color="green",
                    )
                    typer.echo("A few more questions and we're done...\n")
                else:
                    typer.echo(
                        "Sorry, something didn't work out. Please contact us at support@robusta.dev",
                        color="red",
                    )
                    robusta_api_key = ""
        else:
            robusta_api_key = ""

    account_id = str(uuid.uuid4())
    require_eula_approval = False
    if robusta_api_key:  # if Robusta ui sink is defined, take the account id from it
        token = json.loads(base64.b64decode(robusta_api_key))
        account_id = token.get("account_id", account_id)

        sinks_config.append(
            RobustaSinkConfigWrapper(
                robusta_sink=RobustaSinkParams(
                    name="robusta_ui_sink", token=robusta_api_key
                )
            )
        )
        enable_platform_playbooks = True
        require_eula_approval = True

    slack_feedback_heads_up_message: Optional[str] = None
    if slack_integration_configured:
        try:
            slack_feedback_heads_up_message = SlackFeedbackMessagesSender(
                slack_api_key,
                slack_channel,
                account_id,
                debug
            ).schedule_feedback_messages()
        except Exception as e:
            if debug:
                typer.secho(traceback.format_exc())

    if enable_prometheus_stack is None:
        enable_prometheus_stack = typer.confirm(
            "If you haven't installed Prometheus yet, Robusta can install a pre-configured Prometheus. Would you like to do so?"
        )

    if disable_cloud_routing is None:
        disable_cloud_routing = not typer.confirm(
            "Would you like to enable two-way interactivity (e.g. fix-it buttons in Slack) via Robusta's cloud?"
        )
        if not disable_cloud_routing:
            require_eula_approval = True

    if require_eula_approval:
        eula_url = f"{backend_profile.robusta_cloud_api_host}/eula.html"
        typer.echo(
            f"Please read and approve our End User License Agreement: {eula_url}"
        )
        eula_approved = typer.confirm("Do you accept our End User License Agreement?")
        if not eula_approved:
            typer.echo("End User License Agreement rejected. Installation aborted.")
            return

        try:
            requests.get(f"{eula_url}?account_id={account_id}")
        except Exception:
            typer.echo(f"\nEula approval failed: {eula_url}")

    signing_key = str(uuid.uuid4()).replace("_", "")

    values = HelmValues(
        clusterName=cluster_name,
        globalConfig=GlobalConfig(signing_key=signing_key, account_id=account_id),
        sinksConfig=sinks_config,
        enablePrometheusStack=enable_prometheus_stack,
        disableCloudRouting=disable_cloud_routing,
        enablePlatformPlaybooks=enable_platform_playbooks,
        rsa=gen_rsa_pair()
    )

    if is_small_cluster:
        values.set_pod_configs_for_small_clusters()
        values.playbooksPersistentVolumeSize = "128Mi"

    if backend_profile.custom_profile:
        if not values.runner:
            values.runner = {}
        values.runner["additional_env_vars"] = [
            {
                "name": "RELAY_EXTERNAL_ACTIONS_URL",
                "value": backend_profile.robusta_relay_external_actions_url,
            },
            {
                "name": "WEBSOCKET_RELAY_ADDRESS",
                "value": backend_profile.robusta_relay_ws_address,
            },
            {"name": "ROBUSTA_UI_DOMAIN", "value": backend_profile.robusta_ui_domain},
            {
                "name": "ROBUSTA_TELEMETRY_ENDPOINT",
                "value": backend_profile.robusta_telemetry_endpoint
            }
        ]

    write_values_file(output_path, values)

    if robusta_api_key:
        typer.secho(
            f"Finish the Helm install and then login to Robusta UI at {backend_profile.robusta_ui_domain}\n",
            fg="green",
        )

    if slack_feedback_heads_up_message:
        typer.secho(slack_feedback_heads_up_message)


@app.command()
def update_config(
    existing_values: str = typer.Option(
        ...,
        help="Existing values.yaml file name. You can run `helm get values` to get it from the cluster.",
    ),
):
    """
        Update an existing values.yaml file.
        Add RSA key-pair if it doesn't exist
        Add a signing key if it doesn't exist, or replace with a valid one if the key has an invalid format
    """
    with open(existing_values, "r") as existing_values_file:
        values: HelmValues = HelmValues(**yaml.safe_load(existing_values_file))
        if not values.rsa:
            typer.secho("Generating RSA key-pair", fg="green")
            values.rsa = gen_rsa_pair()

        if not values.globalConfig.signing_key:
            typer.secho("Generating signing key", fg="green")
            values.globalConfig.signing_key = str(uuid.uuid4())

        try:
            uuid.UUID(values.globalConfig.signing_key)
        except:
            typer.secho("Invalid signing key. Generating a new one", fg="green")
            values.globalConfig.signing_key = str(uuid.uuid4())

        write_values_file("updated_values.yaml", values)
        typer.secho("Run `helm upgrade robusta -f ./updated_values.yaml`", fg="green")


@app.command()
def playground():
    """Open a python playground - useful when writing playbooks"""
    typer.echo(
        "this command is temporarily disabled due to recent changes to python:3.8-slim"
    )
    # exec_in_robusta_runner("socat readline unix-connect:/tmp/manhole-1")


@app.command()
def version():
    """Show the version of the local robusta-cli"""
    if __version__ == "0.0.0":
        typer.echo("running with development version from git")
    else:
        typer.echo(f"version {__version__}")


@app.command()
def demo():
    """Deliberately deploy a crashing pod to kubernetes so you can test robusta's response"""
    CRASHPOD_YAML = "https://gist.githubusercontent.com/robusta-lab/283609047306dc1f05cf59806ade30b6/raw/crashpod.yaml"
    log_title("Deploying a crashing pod to kubernetes...")
    subprocess.check_call(f"kubectl apply -f {CRASHPOD_YAML}", shell=True)
    log_title(
        "In ~30 seconds you should receive a slack notification on a crashing pod"
    )
    time.sleep(60)
    subprocess.check_call(f"kubectl delete deployment crashpod", shell=True)
    log_title("Done!")


@app.command()
def logs(
    namespace: str = typer.Option(
        None,
        help="Namespace",
    ),
    f: bool = typer.Option(False, "-f", show_default=False, help="Stream runner logs"),
    since: str = typer.Option(
        None, help="Only return logs newer than a relative duration like 5s, 2m, or 3h."
    ),
    tail: int = typer.Option(None, help="Lines of recent log file to display."),
    context: str = typer.Option(
        None,
        help="The name of the kubeconfig context to use"
    ),
):
    """Fetch Robusta runner logs"""
    stream = "-f" if f else ""
    since = f"--since={since}" if since else ""
    tail = f"--tail={tail}" if tail else ""
    context = f"--context={context}" if context else ""
    subprocess.check_call(
        f"kubectl logs {stream} {namespace_to_kubectl(namespace)} deployment/robusta-runner -c runner {since} {tail} {context}",
        shell=True,
    )


if __name__ == "__main__":
    app()



from hikaru import DiffDetail, HikaruBase
from typing import List


# TODO: filter out all the managed fields crap
def is_matching_diff(diff_detail: DiffDetail, fields_to_monitor: List[str]) -> bool:
    return any(
        substring in diff_detail.formatted_path for substring in fields_to_monitor
    )


def duplicate_without_fields(obj: HikaruBase, omitted_fields: List[str]):
    """
    Duplicate a hikaru object, omitting the specified fields

    This is useful when you want to compare two versions of an object and first "cleanup" fields that shouldn't be
    compared.

    :param HikaruBase obj: A kubernetes object
    :param List[str] omitted_fields: List of fields to be omitted. Field name format should be '.' separated
                                     For example: ["status", "metadata.generation"]
    """
    if obj is None:
        return None

    duplication = obj.dup()

    for field_name in omitted_fields:
        field_parts = field_name.split(".")
        try:
            if len(field_parts) > 1:
                parent_obj = duplication.object_at_path(field_parts[:-1])
            else:
                parent_obj = duplication

            setattr(parent_obj, field_parts[-1], None)
        except Exception:
            pass  # in case the field doesn't exist on this object

    return duplication



import inspect
from typing import Callable, Optional, Dict, Type

from pydantic.main import BaseModel

from ..model.base_params import ActionParams
from ..model.events import ExecutionBaseEvent, ExecutionEventBaseParams
from ...utils.decorators import doublewrap


class NotAnActionException(Exception):
    def __init__(self, obj):
        super(NotAnActionException, self).__init__(f"{obj}  is not a playbook action")


@doublewrap
def action(func: Callable):
    """
    Decorator to mark functions as playbook actions
    """
    setattr(func, "_action_name", func.__name__)
    return func


class Action:
    def __init__(
        self,
        func: Callable,
    ):
        if not self.is_action(func):
            raise NotAnActionException(func)

        self.action_name = func.__name__
        self.func = func
        self.event_type = self.__get_action_event_type(func)
        self.params_type = self.__get_action_params_type(func)
        self.from_params_func = None
        self.from_params_parameter_class = None
        if vars(self.event_type).get(
            "from_params"
        ):  # execution event has 'from_params'
            self.from_params_func = getattr(self.event_type, "from_params")
            from_params_signature = inspect.signature(self.from_params_func)
            self.from_params_parameter_class = list(
                from_params_signature.parameters.values()
            )[0].annotation

    @staticmethod
    def is_action(func):
        return (
            inspect.isfunction(func) and getattr(func, "_action_name", None) is not None
        )

    @staticmethod
    def __get_action_event_type(func: Callable):
        """
        Returns the event_type of a playbook action.
        E.g. given an action like:

            @action
            def some_playbook(event: PodEvent, params: MyPlaybookParams):
                pass

        This function returns the class PodEvent
        """
        func_params = iter(inspect.signature(func).parameters.values())
        event_type = next(func_params).annotation
        if not issubclass(event_type, ExecutionBaseEvent):
            raise Exception(
                f"Illegal action first parameter {event_type}. Must extend ExecutionBaseEvent"
            )
        return event_type

    @staticmethod
    def __get_action_params_type(func: Callable):
        """
        Returns the parameters class for a playbook action or None if the action has no parameters
        E.g. given an action like:

            @action
            def some_playbook(event: PodEvent, params: MyPlaybookParams):
                pass

        This function returns the class MyPlaybookParams
        """
        func_params = iter(inspect.signature(func).parameters.values())
        next(func_params)  # skip the event parameter
        action_params = next(func_params, None)
        if not action_params:
            return None
        params_cls = action_params.annotation
        if not issubclass(params_cls, BaseModel):
            raise Exception(
                f"Illegal action second parameter {params_cls}. Action params must extend BaseModel"
            )
        return params_cls


class ActionsRegistry:
    _actions: Dict[str, Action] = {}

    def add_action(self, func: Callable):
        self._actions[func.__name__] = Action(func)

    def get_action(self, action_name: str) -> Optional[Action]:
        return self._actions.get(action_name)

    def get_external_actions(
        self,
    ) -> [(str, Type[ExecutionEventBaseParams], Optional[Type[BaseModel]])]:
        """Should be used to prepare calling schema for each action"""
        return [
            (
                action_def.action_name,
                action_def.from_params_parameter_class,
                action_def.params_type,
            )
            for action_def in self._actions.values()
            if action_def.from_params_func
        ]



import uuid

from typing import List

from .msteams_elements.msteams_text_block import MsTeamsTextBlock
from .msteams_elements.msteams_action import MsTeamsAction
from .msteams_elements.msteams_column import MsTeamsColumn
from .msteams_elements.msteams_container import MsTeamsContainer

from ...core.reporting import FileBlock

'''
there are 3 elements for each text file":
1. row that contains 'open file-name' , and 'close file-name' buttons. 
   the open is visible and the close is invisible at init stage
2. container that contains the last lines of the file - invisible at the init stage
3. row that contains 'close file-name' buttons, invisible at init stage

when a open button is pressed for specific file, the close file for the current file become visible, 
and also the container presenting the file text. for all the other files, the 'open file-name' buttons 
become visible and all the other buttons become invisible

no point in creating object for each file since it will only hold a few variables such as the file name
 and keys and the logic is mainly to create the columns and text containers for all the files combine
'''


class MsTeamsAdaptiveCardFilesText:

    def __init__(self):
        self.open_key_list = []
        self.close_start_key_list = []
        self.close_end_key_list = []
        self.text_file_presentation_key_list = []

        self.open_text_list = []
        self.close_start_text_list = []
        self.close_end_text_list = []
        self.text_file_presentation_list = []

        self.action_open_text_list = []
        self.action_close_start_text_list = []
        self.action_close_end_text_list = []

        self.file_name_list = []

        self.text_map_and_single_text_lines_list = []

    def create_files_for_presentation(self, file_blocks: List[FileBlock]) -> List[map]:
        file_content_list = []

        for file_block in file_blocks:
            if not self.__is_txt_file(file_block.filename):
                continue
            self.__create_new_keys()
            self.file_name_list.append(file_block.filename)
            file_content_list.append(file_block.contents)

        if len(self.open_key_list) == 0:
            return []

        for index in range(len(self.open_key_list)):
            self.__manage_blocks_for_single_file(index, self.file_name_list[index], file_content_list[index])
        return self.__manage_all_text_to_send()

    def get_text_files_containers_list(self):
        return self.text_map_and_single_text_lines_list

    def __create_new_keys(self):
        self.open_key_list.append(str(uuid.uuid4()))
        self.close_start_key_list.append(str(uuid.uuid4()))
        self.close_end_key_list.append(str(uuid.uuid4()))
        self.text_file_presentation_key_list.append(str(uuid.uuid4()))

    def __manage_blocks_for_single_file(self, index, file_name : str, content : bytes):
        open_text_action = self.__action(index, open=True, title='press to open')
        close_text_action = self.__action(index, open=False, title='press to close')

        open_text = MsTeamsTextBlock('***open ' + file_name + '***', is_subtle=False)
        close_start = MsTeamsTextBlock('***close ' + file_name + '***', is_subtle=False)
        close_end = MsTeamsTextBlock('***close ' + file_name + '***', is_subtle=False)

        self.open_text_list.append(open_text)
        self.close_start_text_list.append(close_start)
        self.close_end_text_list.append(close_end)

        self.action_open_text_list.append(open_text_action)
        self.action_close_start_text_list.append(close_text_action)
        self.action_close_end_text_list.append(close_text_action)

        self.text_file_presentation_list.append(self.__present_text_file_block(self.text_file_presentation_key_list[index], content.decode('utf-8')))

    def __manage_all_text_to_send(self):
        top_column_set = MsTeamsColumn()
        bottom_column_set = MsTeamsColumn()
        for index in range(len(self.open_text_list)):
            top_column_set.add_column(key=self.open_key_list[index],
                                      items=[self.open_text_list[index]],
                                      action= self.action_open_text_list[index])

            top_column_set.add_column(is_visible=False,
                                      key=self.close_start_key_list[index],
                                      items=[self.close_start_text_list[index]],
                                      action=self.action_close_start_text_list[index])

            # spaces between files
            top_column_set.add_column(items=[MsTeamsTextBlock(' ')])
            top_column_set.add_column(items=[MsTeamsTextBlock(' ')])

            bottom_column_set.add_column(is_visible=False,
                                         key=self.close_end_key_list[index],
                                         items=[self.close_end_text_list[index]],
                                         action=self.action_close_end_text_list[index])
        
        list_to_return = [top_column_set]
        list_to_return += self.text_file_presentation_list
        list_to_return.append(bottom_column_set)

        return list_to_return
    
    def __action(self, index, open: bool, title: str) -> map:
        visible_elements_map = {False: [], True: []}
        curr_key = self.open_key_list[index]
        for key in self.open_key_list:
            visible = (not open) or (curr_key != key)
            visible_elements_map[visible].append(key)

        curr_key = self.close_start_key_list[index]
        for key in self.close_start_key_list:
            visible = open and (curr_key == key)
            visible_elements_map[visible].append(key)

        curr_key = self.close_end_key_list[index]
        for key in self.close_end_key_list:
            visible = open and (curr_key == key)
            visible_elements_map[visible].append(key)

        curr_key = self.text_file_presentation_key_list[index]
        for key in self.text_file_presentation_key_list:
            visible = open and (curr_key == key)
            visible_elements_map[visible].append(key)

        return MsTeamsAction(title, visible_keys=visible_elements_map[True], invisible_keys=visible_elements_map[False])

    # there is a limit to the number of letters you can write - dont know what it is !!!
    # /t doesn't work so we need to simulate spaces (which are trimmed so we use '. . . ')
    def __present_text_file_block(self, key: str, text: str):
        text_lines_list = []
        new_text = text.replace('\t', '. . . ')

        for line in new_text.split('\n'):
            text_lines_list.append(line + '\n\n')

        # will be completed later
        text_block = MsTeamsTextBlock('', wrap=True, weight='bolder', is_visible=True)
        self.text_map_and_single_text_lines_list.append([text_block, text_lines_list])
        return MsTeamsContainer(key=key, elements=[text_block])

    @classmethod
    def __is_txt_file(cls, file_name: str) -> bool:
        txt_suffixes = ['.txt', '.json', '.yaml', '.log']
        for prefix in txt_suffixes:
            if file_name.lower().endswith(prefix):
                return True
        return False



import logging
from typing import Union, cast, List, Callable
from pydantic import BaseModel

from ...utils.function_hashes import action_hash
from .playbook_scheduler_manager import PlaybooksSchedulerManager
from .trigger import ScheduledTrigger
from ...model.playbook_definition import PlaybookDefinition
from ...model.playbook_action import PlaybookAction
from .event import ScheduledExecutionEvent
from ...core.playbooks.playbooks_event_handler import PlaybooksEventHandler
from ...core.schedule.model import (
    SchedulingInfo,
    ScheduledJob,
    FixedDelayRepeat,
    DynamicDelayRepeat,
    JobState,
)
from ...core.schedule.scheduler import Scheduler

SCHEDULED_INTEGRATION_TASK = "scheduled_integration_task"


class ScheduledIntegrationParams(BaseModel):
    action_func_name: str
    action_params: dict = None
    named_sinks: List[str] = []


class PlaybooksSchedulerManagerImpl(PlaybooksSchedulerManager):
    def __init__(self, event_handler: PlaybooksEventHandler):
        self.event_handler = event_handler
        self.scheduler = Scheduler()
        self.scheduler.register_task(
            SCHEDULED_INTEGRATION_TASK, self.__run_scheduled_task
        )
        self.scheduler.init_scheduler()

    def schedule_playbook(
        self,
        action_name: str,
        playbook_id: str,
        scheduling_params: Union[FixedDelayRepeat, DynamicDelayRepeat],
        named_sinks: List[str],
        action_params=None,
        job_state: JobState = JobState(),
        replace_existing: bool = False,
        standalone_task: bool = False,
    ):
        integration_params = ScheduledIntegrationParams(
            action_func_name=action_name,
            action_params=action_params,
            named_sinks=named_sinks,
        )
        job = ScheduledJob(
            job_id=playbook_id,
            runnable_name=SCHEDULED_INTEGRATION_TASK,
            runnable_params=integration_params.dict(),
            state=job_state,
            scheduling_params=scheduling_params,
            replace_existing=replace_existing,
            standalone_task=standalone_task,
        )
        self.scheduler.schedule_job(job)

    def schedule_action(
        self,
        action_func: Callable,
        task_id: str,
        scheduling_params: Union[FixedDelayRepeat, DynamicDelayRepeat],
        named_sinks: List[str],
        action_params=None,
        job_state: JobState = JobState(),
        replace_existing: bool = False,
        standalone_task: bool = False,
    ):
        playbook_id = action_hash(
            action_func,
            action_params,
            {"key": task_id},
        )
        self.schedule_playbook(
            action_func.__name__,
            playbook_id=playbook_id,
            scheduling_params=scheduling_params,
            named_sinks=named_sinks,
            action_params=action_params,
            replace_existing=replace_existing,
            standalone_task=standalone_task,
        )

    def update(self, playbooks: List[PlaybookDefinition]):
        playbook_ids = set(playbook.get_id() for playbook in playbooks)
        self.__unschedule_deleted_playbooks(playbook_ids)

        for playbook in playbooks:
            if not self.scheduler.is_scheduled(playbook.get_id()):

                # For scheduling simplicity, support only a single trigger and a single action
                if len(playbook.triggers) != 1 or len(playbook.get_actions()) != 1:
                    msg = "Illegal scheduled playbook. Must be a single trigger and a single action"
                    logging.error(msg)
                    raise Exception(msg)

                playbook_trigger: ScheduledTrigger = cast(
                    ScheduledTrigger, playbook.triggers[0].get()
                )
                playbook_action: PlaybookAction = playbook.get_actions()[0]
                self.schedule_playbook(
                    action_name=playbook_action.action_name,
                    playbook_id=playbook.get_id(),
                    scheduling_params=playbook_trigger.get_params(),
                    named_sinks=playbook.sinks,
                    action_params=playbook_action.action_params,
                )

    def __run_scheduled_task(
        self, runnable_params: dict, schedule_info: SchedulingInfo
    ):
        scheduled_params = ScheduledIntegrationParams(**runnable_params)

        action = PlaybookAction(
            action_name=scheduled_params.action_func_name,
            action_params=scheduled_params.action_params,
        )
        execution_event = ScheduledExecutionEvent(
            recurrence=schedule_info.execution_count,
            named_sinks=scheduled_params.named_sinks,
        )
        self.event_handler.run_actions(execution_event, [action])

    def __unschedule_deleted_playbooks(self, active_playbook_ids: set):
        for job in self.scheduler.list_scheduled_jobs():
            if job.standalone_task:
                continue  # standalone tasks shouldn't be removed on reload
            if job.job_id not in active_playbook_ids:
                logging.info(f"unscheduling deleted playbook {job.job_id}")
                self.scheduler.unschedule_job(job.job_id)



from robusta.api import *

from collections import defaultdict, namedtuple

import pygal
from pygal.style import DarkColorizedStyle as ChosenStyle
from prometheus_api_client import PrometheusConnect

from string import Template
from datetime import datetime, timedelta
import humanize


def __prepare_promql_query(provided_labels: Dict[Any, Any], promql_query_template: str) -> str:
    labels = defaultdict(lambda: "<missing>")
    labels.update(provided_labels)
    template = Template(promql_query_template)
    promql_query = template.safe_substitute(labels)
    return promql_query


def get_node_internal_ip(node: Node) -> str:
    internal_ip = next(
        addr.address
        for addr in node.status.addresses
        if addr.type == "InternalIP"
    )
    return internal_ip


def create_chart_from_prometheus_query(
        prometheus_base_url: str,
        promql_query: str,
        starts_at: datetime,
        include_x_axis: bool,
        graph_duration_minutes: int,
        chart_title: Optional[str] = None,
        values_format: Optional[ChartValuesFormat] = None
):
    if not prometheus_base_url:
        prometheus_base_url = PrometheusDiscovery.find_prometheus_url()
    prom = PrometheusConnect(url=prometheus_base_url, disable_ssl=True)
    end_time = datetime.now(tz=starts_at.tzinfo)
    alert_duration = end_time - starts_at
    graph_duration = max(alert_duration, timedelta(minutes=graph_duration_minutes))
    start_time = end_time - graph_duration
    resolution = 250  # 250 is used in Prometheus web client in /graph and looks good
    increment = max(graph_duration.total_seconds() / resolution, 1.0)
    result = prom.custom_query_range(
        promql_query,
        start_time,
        end_time,
        increment,
        {"timeout": PROMETHEUS_REQUEST_TIMEOUT_SECONDS},
    )

    chart = pygal.XY(
        show_dots=True,
        style=ChosenStyle,
        truncate_legend=15,
        include_x_axis=include_x_axis,
        width=1280,
        height=720
    )

    chart.x_label_rotation = 35
    chart.truncate_label = -1
    chart.x_value_formatter = lambda timestamp: datetime.fromtimestamp(
        timestamp
    ).strftime("%I:%M:%S %p on %d, %b")

    value_formatters = {
        ChartValuesFormat.Plain: lambda val: str(val),
        ChartValuesFormat.Bytes: lambda val: humanize.naturalsize(val, binary=True),
        ChartValuesFormat.Percentage: lambda val: f'{(100*val):.1f}%'
    }
    chart_values_format = values_format if values_format else ChartValuesFormat.Plain
    chart.value_formatter = value_formatters[chart_values_format]

    if chart_title:
        chart.title = chart_title
    else:
        chart.title = promql_query
    # fix a pygal bug which causes infinite loops due to rounding errors with floating points
    for series in result:
        label = "\n".join([v for v in series["metric"].values()])
        values = [
            (timestamp, round(float(val), FLOAT_PRECISION_LIMIT))
            for (timestamp, val) in series["values"]
        ]
        chart.add(label, values)
    return chart


def create_graph_enrichment(
        start_at: datetime,
        labels: Dict[Any, Any],
        promql_query: str,
        prometheus_url: Optional[str],
        graph_duration_minutes: int,
        graph_title: Optional[str],
        chart_values_format: Optional[ChartValuesFormat]) -> FileBlock:
    promql_query = __prepare_promql_query(labels, promql_query)
    chart = create_chart_from_prometheus_query(
        prometheus_url,
        promql_query,
        start_at,
        include_x_axis=True,
        graph_duration_minutes=graph_duration_minutes,
        chart_title=graph_title,
        values_format=chart_values_format
    )
    chart_name = graph_title if graph_title else promql_query
    svg_name = f"{chart_name}.svg"
    return FileBlock(svg_name, chart.render())


def create_resource_enrichment(
    starts_at: datetime,
    labels: Dict[Any, Any],
    resource_type: ResourceChartResourceType,
    item_type: ResourceChartItemType,
    graph_duration_minutes: int,
    prometheus_url: Optional[str] = None,
) -> FileBlock:
    ChartOptions = namedtuple('ChartOptions', ['query', 'values_format'])
    combinations = {
        (ResourceChartResourceType.CPU, ResourceChartItemType.Pod): ChartOptions(
            query='sum(node_namespace_pod_container:container_cpu_usage_seconds_total:sum_irate{namespace="$namespace", pod=~"$pod"})',
            values_format=ChartValuesFormat.Plain
        ),
        (ResourceChartResourceType.CPU, ResourceChartItemType.Node): ChartOptions(
            query='instance:node_cpu_utilisation:rate5m{job="node-exporter", instance=~"$node_internal_ip:[0-9]+", cluster=""} != 0',
            values_format=ChartValuesFormat.Percentage
        ),
        (ResourceChartResourceType.Memory, ResourceChartItemType.Pod): ChartOptions(
            query='sum(container_memory_working_set_bytes{job="kubelet", metrics_path="/metrics/cadvisor", pod=~"$pod", container!="", image!=""})',
            values_format=ChartValuesFormat.Bytes
        ),
        (ResourceChartResourceType.Memory, ResourceChartItemType.Node): ChartOptions(
            query='instance:node_memory_utilisation:ratio{job="node-exporter", instance=~"$node_internal_ip:[0-9]+", cluster=""} != 0',
            values_format=ChartValuesFormat.Percentage
        ),
        (ResourceChartResourceType.Disk, ResourceChartItemType.Pod): None,
        (ResourceChartResourceType.Disk, ResourceChartItemType.Node): ChartOptions(
            query='sum(sort_desc(1 -(max without (mountpoint, fstype) (node_filesystem_avail_bytes{job="node-exporter", fstype!="", instance=~"$node_internal_ip:[0-9]+", cluster=""})/max without (mountpoint, fstype) (node_filesystem_size_bytes{job="node-exporter", fstype!="", instance=~"$node_internal_ip:[0-9]+", cluster=""})) != 0))',
            values_format=ChartValuesFormat.Percentage
        ),
    }
    combination = (resource_type, item_type)
    chosen_combination = combinations[combination]
    if not chosen_combination:
        raise AttributeError(f'The following combination for resource chart is not supported: {combination}')
    values_format_text = 'Utilization' if chosen_combination.values_format == ChartValuesFormat.Percentage else 'Usage'
    graph_enrichment = create_graph_enrichment(
        starts_at,
        labels,
        chosen_combination.query,
        prometheus_url=prometheus_url,
        graph_duration_minutes=graph_duration_minutes,
        graph_title=f'{resource_type.name} {values_format_text} for this {item_type.name.lower()}',
        chart_values_format=chosen_combination.values_format
    )
    return graph_enrichment



import logging
import os
import requests

from ....core.reporting.utils import convert_svg_to_png, is_image, SVG_SUFFIX, PNG_SUFFIX

TELEGRAM_BASE_URL = os.environ.get("TELEGRAM_BASE_URL", "https://api.telegram.org")


class TelegramClient:
    def __init__(self, chat_id: int, bot_token: str):
        self.chat_id = chat_id
        self.bot_token = bot_token

    def send_message(self, message: str, disable_links_preview: bool = True):
        url = f"{TELEGRAM_BASE_URL}/bot{self.bot_token}/sendMessage"
        message_json = {
            "chat_id": self.chat_id,
            "disable_web_page_preview": disable_links_preview,
            "parse_mode": "Markdown",
            "text": message
        }
        response = requests.post(url, json=message_json)

        if response.status_code != 200:
            logging.error(
                f"Failed to send telegram message: chat_id - {self.chat_id} reason - {response.reason} {response.text}"
            )

    def send_file(self, file_name: str, contents: bytes):
        file_type = "Photo" if is_image(file_name) else "Document"
        url = f"{TELEGRAM_BASE_URL}/bot{self.bot_token}/send{file_type}?chat_id={self.chat_id}"
        if file_name.endswith(SVG_SUFFIX):
            contents = convert_svg_to_png(contents)
            file_name = file_name.replace(SVG_SUFFIX, PNG_SUFFIX)

        files = {
            file_type.lower(): (file_name, contents)
        }
        response = requests.post(url, files=files)

        if response.status_code != 200:
            logging.error(
                f"Failed to send telegram file: chat_id - {self.chat_id} reason - {response.reason} {response.text}"
            )




import base64
import json
import subprocess
import traceback
import uuid
from typing import Optional
import click_spinner
from dpath.util import get
import requests
import typer
import yaml
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat, PublicFormat, NoEncryption
from cryptography.hazmat.primitives.asymmetric import rsa
from pydantic import BaseModel

from .backend_profile import backend_profile
from .playbooks_cmd import NAMESPACE_EXPLANATION, get_playbooks_config
from .utils import namespace_to_kubectl

AUTH_SECRET_NAME = "robusta-auth-config-secret"
app = typer.Typer()


class RSAKeyPair(BaseModel):
    prv: str
    pub: str


def gen_rsa_pair() -> RSAKeyPair:
    # generate private/public key pair
    key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    # get public key in OpenSSH format
    public_key = key.public_key().public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)

    # get private key in PEM container format
    pem = key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=NoEncryption()
    )

    return RSAKeyPair(
        pub=public_key.decode('utf-8'),
        prv=pem.decode('utf-8')
    )


def get_existing_auth_config(namespace: str) -> Optional[RSAKeyPair]:
    try:
        secret_content = subprocess.check_output(
            f"kubectl get secret {namespace_to_kubectl(namespace)} {AUTH_SECRET_NAME} -o yaml",
            shell=True,
        )
    except Exception:
        return None

    auth_secret = yaml.safe_load(secret_content)
    return RSAKeyPair(
        prv=base64.b64decode(auth_secret["data"]["prv"]).decode(),
        pub=base64.b64decode(auth_secret["data"]["pub"]).decode()
    )


class TokenDetails(BaseModel):
    pub: str
    account_id: str
    user_id: str
    session_token: str
    enc_key: str
    key_id: str


def store_server_token(token_details: TokenDetails, debug: bool = False) -> bool:
    try:
        response = requests.post(backend_profile.robusta_store_token_url, json=token_details.dict())
        if debug and response.status_code != 201:
            typer.secho(f"Failed to store server token. status-code {response.status_code} text {response.text}")

        return response.status_code == 201
    except Exception as e:
        if debug:
            typer.secho(f"Error trying to store server token. {traceback.format_exc()}")
        return False


@app.command()
def gen_token(
    account_id: str = typer.Option(
        ...,
        help="Robusta account id",
    ),
    user_id: str = typer.Option(
        ...,
        help="User id for which the token is created",
    ),
    session_token: str = typer.Option(
        ...,
        help="User session token. Created for an authenticated user via the Robusta UI",
    ),
    namespace: str = typer.Option(
        None,
        help=NAMESPACE_EXPLANATION,
    ),
    debug: bool = typer.Option(False),
):
    """Generate token required to run actions manually in Robusta UI"""
    typer.echo("connecting to cluster...")
    with click_spinner.spinner():
        auth_config = get_existing_auth_config(namespace)

    if not auth_config:
        typer.secho("\nRSA auth isn't configured. "
                    "Please update Robusta and run `robusta update-config` to configure it. Aborting!", fg="red")
        return

    playbooks_config = get_playbooks_config(namespace)
    active_playbooks_file = playbooks_config["data"]["active_playbooks.yaml"]
    playbooks_config_yaml = yaml.safe_load(active_playbooks_file)
    signing_key = get(playbooks_config_yaml, "global_config/signing_key", default=None)
    if not signing_key:
        typer.secho("signing_key is not defined. Please update Robusta and run `robusta update-config`", fg="red")
        return

    try:
        signing_key = uuid.UUID(signing_key)
    except Exception:
        typer.secho("Bad format for signing_key. Please run `robusta update-config` to generate a new valid"
                    " signing_key for your account.", fg="red")
        return

    client_enc_key = uuid.uuid4()
    server_enc_key = uuid.UUID(int=(signing_key.int ^ client_enc_key.int))
    key_id = str(uuid.uuid4())

    token_response = TokenDetails(
        pub=auth_config.pub,
        account_id=account_id,
        user_id=user_id,
        session_token=session_token,
        enc_key=str(server_enc_key),
        key_id=key_id,
    )
    if not store_server_token(token_response, debug):
        typer.secho("Failed to store server token. Aborting!", fg="red")
        return

    # client response is the same, only with a different enc_key
    token_response.enc_key = str(client_enc_key)

    typer.secho(f"Token created successfully. Submit it in the Robusta UI", fg="green")
    typer.secho(str(base64.b64encode(json.dumps(token_response.json(exclude={"session_token"})).encode("utf-8"))))



from ..sink_base_params import SinkBaseParams
from ..sink_config import SinkConfigBase


class KafkaSinkParams(SinkBaseParams):
    kafka_url: str
    topic: str


class KafkaSinkConfigWrapper(SinkConfigBase):
    kafka_sink: KafkaSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.kafka_sink



from ..sink_config import SinkConfigBase
from ..sink_base_params import SinkBaseParams


class WebhookSinkParams(SinkBaseParams):
    url: str
    size_limit: int = 4096


class WebhookSinkConfigWrapper(SinkConfigBase):
    webhook_sink: WebhookSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.webhook_sink




from typing import List

from .msteams_adaptive_card_files_image import MsTeamsAdaptiveCardFilesImage
from .msteams_adaptive_card_files_text import MsTeamsAdaptiveCardFilesText
from ...core.reporting import FileBlock


class MsTeamsAdaptiveCardFiles():

    def __init__(self):
        self.text_files = MsTeamsAdaptiveCardFilesText()

    # return List of MsTeamsBaseElement - cant return it in the constructor.
    def upload_files(self, file_blocks: List[FileBlock]) -> List[map]:
        image_section_map: map = MsTeamsAdaptiveCardFilesImage().create_files_for_presentation(file_blocks)
        text_files_section_list = self.text_files.create_files_for_presentation(file_blocks)

        if image_section_map:
            image_section_map = [image_section_map]
        else:
            image_section_map = []
        return text_files_section_list + image_section_map

    # return the list of text containers with the list of lines, so later after
    # calculating the length of bytes left in the message, we can put the 
    # lines evenly in each text container so we dont exceed the msg length
    def get_text_files_containers_list(self):
        return self.text_files.get_text_files_containers_list()


import asyncio
import logging
import threading
import time
from watchgod import watch

WAIT_SEC = 2


class FileSystemWatcher:
    def __init__(self, path_to_watch, reload_configuration):
        self.active = True
        self.change_detected = False
        self.path_to_watch = path_to_watch
        self.reload_configuration = reload_configuration
        self.stop_event = asyncio.Event()

        self.watch_thread = threading.Thread(target=self.watch, name="config-watcher")
        self.watch_thread.start()
        self.fs_watch_thread = threading.Thread(target=self.fs_watch, name="fs-watcher")
        self.fs_watch_thread.start()

        logging.info(f"watching dir {path_to_watch} for custom playbooks changes")

    def fs_watch(self):
        for _ in watch(self.path_to_watch, stop_event=self.stop_event):
            self.mark_change()

    def watch(self):
        while self.active:
            time.sleep(WAIT_SEC)

            if self.change_detected:
                time.sleep(
                    WAIT_SEC
                )  # once we detected a change, we wait a safety period to make sure all the changes under this 'bulk' are finished
                self.change_detected = False
                try:
                    self.reload_configuration(self.path_to_watch)
                except Exception as e:  # in case we have an error while trying to reload, we want the watch thread to stay alive
                    logging.exception("failed to reload configuration")

    def stop_watcher(self):
        self.active = False
        self.stop_event.set()

    def mark_change(self):
        self.change_detected = True



import base64
import hashlib
import hmac
import logging
import time
from typing import Optional, Dict, Any
from uuid import UUID
import websocket
import json
import os
from threading import Thread
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from ..core.playbooks.playbooks_event_handler import PlaybooksEventHandler
from ..core.model.env_vars import INCOMING_REQUEST_TIME_WINDOW_SECONDS, RUNNER_VERSION
from ..core.reporting.action_requests import (
    ExternalActionRequest,
    ActionRequestBody,
    sign_action_request, PartialAuth,
)
from ..utils.auth_provider import AuthProvider

WEBSOCKET_RELAY_ADDRESS = os.environ.get(
    "WEBSOCKET_RELAY_ADDRESS", "wss://relay.robusta.dev"
)
CLOUD_ROUTING = json.loads(os.environ.get("CLOUD_ROUTING", "True").lower())
RECEIVER_ENABLE_WEBSOCKET_TRACING = json.loads(
    os.environ.get("RECEIVER_ENABLE_WEBSOCKET_TRACING", "False").lower()
)
INCOMING_WEBSOCKET_RECONNECT_DELAY_SEC = int(
    os.environ.get("INCOMING_WEBSOCKET_RECONNECT_DELAY_SEC", 3)
)


class ActionRequestReceiver:
    def __init__(self, event_handler: PlaybooksEventHandler):
        self.event_handler = event_handler
        self.active = True
        self.account_id = self.event_handler.get_global_config().get("account_id")
        self.cluster_name = self.event_handler.get_global_config().get("cluster_name")
        self.auth_provider = AuthProvider()

        self.ws = websocket.WebSocketApp(
            WEBSOCKET_RELAY_ADDRESS,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
        )

        if not self.account_id or not self.cluster_name:
            logging.error(
                f"Action receiver cannot start. "
                f"Missing required account_id {self.account_id} cluster_name {self.cluster_name}"
            )
            return

        self.start_receiver()

    def start_receiver(self):
        if not CLOUD_ROUTING:
            logging.info(
                "outgoing messages only mode. Incoming event receiver not initialized"
            )
            return

        if WEBSOCKET_RELAY_ADDRESS == "":
            logging.warning("relay address empty. Not initializing relay")
            return

        websocket.enableTrace(RECEIVER_ENABLE_WEBSOCKET_TRACING)
        receiver_thread = Thread(target=self.run_forever)
        receiver_thread.start()

    def run_forever(self):
        logging.info("starting relay receiver")
        while self.active:
            self.ws.run_forever()
            logging.info("relay websocket closed")
            time.sleep(INCOMING_WEBSOCKET_RECONNECT_DELAY_SEC)

    def stop(self):
        logging.info(f"Stopping incoming receiver")
        self.active = False
        self.ws.close()

    @classmethod
    def __sync_response(cls, status_code: int, request_id: str, data) -> Dict:
        return {
            "action": "response",
            "request_id": request_id,
            "status_code": status_code,
            "data": data
        }

    def __exec_external_request(
        self, action_request: ExternalActionRequest, validate_timestamp: bool
    ):
        sync_response = action_request.request_id != ""  # if request_id is set, we need to write back the response
        if not self.__validate_request(action_request, validate_timestamp):
            logging.error(f"Failed to validate action request {action_request}")
            if sync_response:
                self.ws.send(data=json.dumps(self.__sync_response(401, action_request.request_id, "")))
            return

        logging.info(f"got callback `{action_request.body.action_name}` {action_request.body.action_params}")
        response = self.event_handler.run_external_action(
            action_request.body.action_name,
            action_request.body.action_params,
            action_request.body.sinks,
            sync_response,
            action_request.no_sinks,
        )

        if sync_response:
            http_code = 200 if response.get("success") else 500
            self.ws.send(data=json.dumps(self.__sync_response(http_code, action_request.request_id, response)))

    def on_message(self, ws, message):
        # TODO: use typed pydantic classes here?
        logging.debug(f"received incoming message {message}")
        incoming_event = json.loads(message)
        actions = incoming_event.get("actions", None)
        if actions:  # this is slack callback format
            # slack callbacks have a list of 'actions'. Within each action there a 'value' field,
            # which container the actual action details we need to run.
            # This wrapper format is part of the slack API, and cannot be changed by us.
            for action in actions:
                try:
                    self.__exec_external_request(
                        ExternalActionRequest.parse_raw(action["value"]), False
                    )
                except Exception:
                    logging.error(
                        f"Failed to run incoming event {incoming_event}", exc_info=True
                    )
        else:  # assume it's ActionRequest format
            try:
                self.__exec_external_request(
                    ExternalActionRequest(**incoming_event), True
                )
            except Exception:
                logging.error(
                    f"Failed to run incoming event {incoming_event}", exc_info=True
                )

    def on_error(self, ws, error):
        logging.info(f"Relay websocket error: {error}")

    def on_open(self, ws):
        account_id = self.event_handler.get_global_config().get("account_id")
        cluster_name = self.event_handler.get_global_config().get("cluster_name")
        open_payload = {
            "action": "auth",
            "account_id": account_id,
            "cluster_name": cluster_name,
            "version": RUNNER_VERSION,
        }
        logging.info(
            f"connecting to server as account_id={account_id}; cluster_name={cluster_name}"
        )
        ws.send(json.dumps(open_payload))

    def __validate_request(self, action_request: ExternalActionRequest, validate_timestamp: bool) -> bool:
        """
            Two auth protocols are supported:
            1. signature - Signing the body using the signing_key should match the signature
            2. partial keys auth - using partial_auth_a and partial_auth_b
               Each partial auth should be decrypted using the private key (rsa private key).
               The content should have 2 items:
               - key
               - body hash
               The operation key_a XOR key_b should be equal to the signing_key

            If both protocols are present, we only check the signature
        """
        if validate_timestamp and (
            time.time() - action_request.body.timestamp
            > INCOMING_REQUEST_TIME_WINDOW_SECONDS
        ):
            logging.error(
                f"Rejecting incoming request because it's too old. Cannot verify request {action_request}"
            )
            return False

        signing_key = self.event_handler.get_global_config().get("signing_key")
        body = action_request.body
        if not signing_key:
            logging.error(f"Signing key not available. Cannot verify request {body}")
            return False

        # First auth protocol option, based on signature only
        signature = action_request.signature
        if signature:
            generated_signature = sign_action_request(body, signing_key)
            return hmac.compare_digest(generated_signature, signature)

        # Second auth protocol option, based on public key
        partial_auth_a = action_request.partial_auth_a
        partial_auth_b = action_request.partial_auth_b
        if not partial_auth_a or not partial_auth_b:
            logging.error(f"Insufficient authentication data. Cannot verify request {body}")
            return False

        private_key = self.auth_provider.get_private_rsa_key()
        if not private_key:
            logging.error(f"Private RSA key missing. Cannot validate request for {body}")
            return False

        a_valid, key_a = self.__extract_key_and_validate(partial_auth_a, private_key, body)
        b_valid, key_b = self.__extract_key_and_validate(partial_auth_b, private_key, body)

        if not a_valid or not b_valid:
            logging.error(f"Cloud not validate partial auth for {body}")
            return False

        try:
            signing_key_uuid = UUID(signing_key)
        except Exception:
            logging.error(f"Wrong signing key format. Cannot validate parital auth for {body}")
            return False

        if (key_a.int ^ key_b.int) != signing_key_uuid.int:
            logging.error(f"Partial auth keys combination mismatch for {body}")
            return False

        return True

    @classmethod
    def __extract_key_and_validate(
            cls,
            encrypted: str,
            private_key: RSAPrivateKey,
            body: ActionRequestBody
    ) -> (bool, Optional[UUID]):
        try:
            plain = private_key.decrypt(
                base64.b64decode(encrypted.encode("utf-8")),
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None
                )
            )
            auth = PartialAuth(**json.loads(plain.decode("utf-8")))
            body_string = body.json(exclude_none=True, sort_keys=True, separators=(',', ':')).encode("utf-8")
            body_hash = f"v0={hashlib.sha256(body_string).hexdigest()}"
            return hmac.compare_digest(body_hash, auth.hash), auth.key
        except Exception:
            logging.error("Error validating partial auth data", exc_info=True)
            return False, None



from .slack_sink_params import SlackSinkConfigWrapper
from ....integrations.slack import SlackSender
from ...reporting.base import Finding
from ..sink_base import SinkBase


class SlackSink(SinkBase):
    def __init__(
        self,
        sink_config: SlackSinkConfigWrapper,
        registry
    ):
        super().__init__(sink_config.slack_sink, registry)
        self.slack_channel = sink_config.slack_sink.slack_channel
        self.api_key = sink_config.slack_sink.api_key
        self.slack_sender = SlackSender(
            self.api_key, self.account_id, self.cluster_name, self.signing_key
        )

    def write_finding(self, finding: Finding, platform_enabled: bool):
        self.slack_sender.send_finding_to_slack(
            finding, self.slack_channel, self.sink_name, platform_enabled
        )



from ..sink_config import SinkConfigBase
from ..sink_base_params import SinkBaseParams


class DataDogSinkParams(SinkBaseParams):
    api_key: str


class DataDogSinkConfigWrapper(SinkConfigBase):
    datadog_sink: DataDogSinkParams

    def get_params(self) -> SinkBaseParams:
        return self.datadog_sink

